#!/usr/bin/env python
# coding: utf-8

import numpy as np
import mpmath
from scipy import stats
from scipy import spatial # for QHull
from scipy import special # for QHull outside integration

from . import sball

import quadpy

from collections import namedtuple

mpmath.mp.dps = 325 # to cover anything that double-presigion float can handle

# maximum radius, where norm.pdf() wasn't zero
# -38.575500173381374935388521407730877399444580.....
# don't ask me what the magic python use to distinguish
# digits after double precision
max_R_ever = 37

#č jako sůl potřebujem statefull třidu,
#č která by schovávala vnitřní implementaciju,
#č která by nabízela jednotné rozhraní pro ostatní kód WellMet
#č a byla by přiměřeně kompatibilní s ConvexHull ze scipy.spatial
#č Ze scipy bych viděl podporu atributů .points, .npoints, .equations
#č Určitě bych chtěl funkce .update() a .is_outside(sample)
#č Atribute .space by ukazoval, v jakém prostoru konvexní obál je vytvořen
# GBall, BrickHull, DirectHull, CompleteHull and QHull has been implemented


def calculate_brick_complement_probability(mins, maxs):
    "For Gaussian space"
    #č na začátku nastavíme obyčejnou 1
    #č v cyklu bude nasobit delkami 
    #č podle jednotlivých souřadnic
    #č a typ se změní na mpf (hodně přesná aritmetika)
    volume = 1
    
    # ncdf() do not support matrix input
    for xmin, xmax in zip(mins, maxs):
        volume *= mpmath.ncdf(xmax) - mpmath.ncdf(xmin)
    
    # print test
#    if len(mins) == 2:
#        min_x, min_y = mins
#        max_x, max_y = maxs
#        test_volume = 0
#        test_volume += np.sum(stats.norm.cdf(mins)) 
#        test_volume += np.sum(stats.norm.sf(maxs))
#        test_volume -= stats.norm.cdf(min_x) * stats.norm.cdf(min_y)
#        test_volume -= stats.norm.cdf(min_x) * stats.norm.sf(max_y)
#        test_volume -= stats.norm.sf(max_x) * stats.norm.cdf(min_y)
#        test_volume -= stats.norm.sf(max_x) * stats.norm.sf(max_y)
#        print("test_volume:", test_volume)  
        
    return float(1-volume)
    
# inspired by 
# https://math.stackexchange.com/questions/192610/how-to-build-a-orthogonal-basis-from-a-vector/979013
# most interesting from there:
# """Pick a vector. WLOG, you chose $(x_1,x_2,x_3,x_4)$. 
# Now write it as a quaternion: $$x_1+ix_2+jx_3+kx_4$$ 
# Then, since multiplication by $i,j,k$ 
# rotates this vector $90^0$ across the various axes of our 4D space,
# the following three vectors make your initial choice of vector into an orthonormal basis: 
# $$i(x_1+ix_2+jx_3+kx_4)=ix_1-x_2+kx_3-jx_4\mapsto (-x_2,x_1,-x_4,x_3)$$ 
# $$j(x_1+ix_2+jx_3+kx_4)=jx_1-kx_2-x_3+ix_4\mapsto (-x_3,x_4,x_1,-x_2)$$
# $$k(x_1+ix_2+jx_3+kx_4)=kx_1+jx_2-ix_3-x_4\mapsto (-x_4,-x_3,x_2,x_1)$$"""
def get_orth_basis(vector):
    dim = len(vector)
    if dim == 2:
        x, y = vector
        return np.array([[x, y],[-y, x]])
    elif dim == 4:
        x_1, x_2, x_3, x_4 = vector
        return np.array([
            [x_1, x_2, x_3, x_4],
            [-x_2, x_1, -x_4, x_3],
            [-x_3, x_4, x_1, -x_2],
            [-x_4, -x_3, x_2, x_1]])
    else:
        random_basis = np.random.random((dim, dim))
        #č QR rozklad jede po sloupcich
        random_basis[:,0] = vector
        #č co jsem viděl, numpy matici Q normalizuje
        #č a první sloupec zůstavá (skoro) tím samým, co byl před tím
        Q, __R = np.linalg.qr(random_basis)
        # yep. Numpy sometimes inverts sign of the first Q matrix vector
        Q[:,0] = vector
        return Q.T
    
        
        
#ё рельса
#č nepodařílo se mi nějak rozumně zobecnit pro libovolný prostor
#č takže Gauss
#č (každá třida implementuje zvlášť)

def shot(hull, ns, use_MC=False):
    if hull.space == 'G':
        to_fire = np.nanargmax(hull.b)
        r = -hull.b[to_fire]
        a = hull.A[to_fire]
        
        if use_MC:
            fire_from = stats.norm.sf(r)
            t = np.linspace(fire_from, 0, ns, endpoint=False)
            t = stats.norm.isf(t)
        else:
            if r < max_R_ever:
                R = max_R_ever
            else:
                R = r + 10
            t = np.linspace(r, R, ns, endpoint=True)
            
        fire_G = t.reshape(-1,1) @ a.reshape(1,-1)
        return hull.sample.f_model.new_sample(fire_G, space='G')


def fire(hull, ns, use_MC=False):
    if hull.space == 'G':
        to_fire = np.nanargmax(hull.b)
        r = -hull.b[to_fire]
        a = hull.A[to_fire]
        
        orth_nodes_T = np.random.randn(len(a), ns) # len(a) == ndim
        orth_basis = get_orth_basis(a)
        
        if use_MC:
            fire_from = stats.norm.sf(r)
            t = np.linspace(fire_from, 0, ns, endpoint=False)
            orth_nodes_T[0] = stats.norm.isf(t)
        else:
            if r < max_R_ever:
                R = max_R_ever
            else:
                R = r + 10
            orth_nodes_T[0] = np.linspace(r, R, ns, endpoint=True)
            
        fire_G = (orth_basis.T @ orth_nodes_T).T
        return hull.sample.f_model.new_sample(fire_G, space='G')

#
# Common orth functions for DirectHull, CompleteHull and QHull
#

# (even private function can be shared :D )
def _orth_helper(hull):
    # supposed hull.space == 'G'
    hull._update()
    
    
    ndim = hull.sample.nvar
    #č je to špatné z hlediska testování,
    #č ale nevím jak z toho.
    #č pro 2D musím udělat vyjimku z "předposlední"
    if ndim > 2:
        orth_basis = _get_nD_orth_basis(hull)
    else:
        to_fire = np.nanargmax(hull.b)
        a = hull.A[to_fire]
        orth_basis = get_orth_basis(a)
    
    #č musí tam bejt G coordinates
    A = orth_basis @ hull.points.T
    bp = np.nanmax(A, axis=1)
    bm = np.nanmin(A, axis=1)
    
    return orth_basis, bp, bm
    
    

def _get_nD_orth_basis(hull):
    # supposed hull.space == 'G'
    # supposed hull._update()
    
    #č bůhví, jestli obálka negeneruje A a b-čko dynamicky
    b = hull.b
    A = hull.A 
    
    return _get_nD_orth_basis_from_normals(A, b)
    
    
def _get_nD_orth_basis_from_normals(A, b):
    _nplates, ndim = A.shape
    
    #č pro 2D "nastav první a pak po veškerem zbytku 
    #č ještě ten předposlední vektor" špatně funguje
    assert ndim > 2
    
    to_fire = np.nanargmax(b)
    a = A[to_fire]
    
    
    #č QR rozklad jede po sloupcich
    basis_T = np.empty((ndim, ndim)) 
    basis_T[:,0] = a
    
    #č skalarní součín
    cos = A @ a
    bcos = b * (1 + np.abs(cos))
    
    for i in range(1, ndim-2):
        
        to_fire = np.nanargmax(bcos)
        next_vector = A[to_fire]
        
        #č QR rozklad komplikuje život transponovanim matic
        basis_T[:,i] = next_vector
        basis_T[:, :i+1], __R = np.linalg.qr(basis_T[:, :i+1])
        
        next_orth_vector = basis_T[:,i]
        
        #č skalarní součín
        cos = A @ next_orth_vector
        bcos *= (1 + np.abs(cos))
        
    
    to_fire = np.nanargmax(bcos)
    #č předposlední vektor
    next_vector = A[to_fire]
    basis_T[:,ndim-2] = next_vector
    #č a už ready. Poslední vektor je určen předchozími
    basis_T, __R = np.linalg.qr(basis_T)
    
    #č QR rozklad jede po sloupcich
    #č co jsem viděl, numpy matici Q normalizuje
    #č a první sloupec zůstavá (skoro) tím samým, co byl před tím
    #Q, __R = np.linalg.qr(basis.T)
    # yep. Numpy sometimes inverts sign of the first Q matrix vector
    #č ale na znáky kašleme. To je důležité pro kandidaty,
    #č zde nám ale jde o basis
    orth_basis = basis_T.T
    
    return orth_basis
    
    
def get_orth_outside(hull):
    if hull.space == 'G' :
        # hull._update() will perform _orth_helper()
        A, bp, bm = hull._orth_helper()
        return calculate_brick_complement_probability(bm, bp)
    else:
        #č když prostor není G, můžeme sice něco odvodit od s-ball 
        #č ale nechť se s tím pará GHull, vrátíme nulu.
        return 0

def get_orth_equations(hull):
    # hull._update() will perform _orth_helper()
    direct_plan, bp, bm = hull._orth_helper()

    A = np.vstack((-direct_plan, direct_plan))
    b = np.concatenate((bm, -bp))
    return np.hstack((A,b[:,None]))     


#
# Common 2FORM functions for DirectHull, CompleteHull and QHull
#

# (even private function can be shared :D )
def _2FORM_helper(hull):
    # supposed hull.space == 'G'
    hull._update()
    to_fire = np.nanargmax(hull.b)
    a = hull.A[to_fire]
    
    #č musí tam bejt G coordinates
    # we'll get 1D npoints-sized negative b array
    x = a @ hull.points.T
    # here x == _b array has another meaning than hull.b
    # where in hull.b we search for least distanced hyperplane
    # here, in x, we search for a valid b_backward hyperplane offset
    # It can be seen as transformation of the ED to 1D a-projection.
    # There evidently exist min(x) and max(x)
    # such that every single x value lies inside (between)
    # max(x) (forward) value should be corresponding to r, to -hull.b
    # So, we need to find min(x) (backward) value
    # x_forward == max(x) == -b == r == max(a @ points.T) == -max(hull.b) == min(-hull.b)
    # x_backward == min(x) == max(-a @ points.T)
    x_forward = np.max(x)
    x_backward = np.min(x)
    #print(-np.max(hull.b), x_forward) 
    return x_backward, x_forward, a


def get_2FORM_outside(hull):
    if hull.space == 'G':
        # hull._update() will perform _FORM_helper()
        x_backward, x_forward, __a = hull._2FORM_helper()
        return stats.norm.cdf(x_backward) + stats.norm.sf(x_forward)
    else:
        #č když prostor není G, můžeme sice něco odvodit od s-ball 
        #č ale nechť se s tím pará GHull, vrátíme nulu.
        return 0
            
def get_2FORM_equations(hull):
    # hull._update() will perform _2FORM_helper()
    x_backward, x_forward, a = hull._2FORM_helper()
    A = np.vstack((-a, a))
    b = np.array([[x_backward],[-x_forward]])
    return np.hstack((A,b))  
    
def get_bp_bm(A, points):
    B = A @ points.T
    bp = np.nanmax(B, axis=1)
    bm = np.nanmin(B, axis=1)
    return bp, bm
    

#č jistě musíme mít nějaký zbytečný kus kódu
#č třida jen pro formu, jen tak na hračku
#č když tečíčky jsou dále jak nejvzdálenější vzorek (bod),
#č tak řekneme, že jsou totálně mimo!
class GBall:
    def __init__(hull, sample):
        hull.sample = sample
        hull.space = 'G' #оӵ малы? Кинлы со кулэ?
    
    @property
    def points(hull):
        return hull.sample.G
        
    @property
    def npoints(hull):
        return len(hull.sample)
        
    @property
    def nsimplex(hull):
        return 1
        
    #def update(hull): pass
    
    def is_inside(hull, nodes):
        R2_hull = np.nanmax(np.sum(np.square(hull.sample.G), axis=1))
        
        R2_nodes = np.sum(np.square(nodes.G), axis=1)
        return R2_nodes < R2_hull 

    def is_outside(hull, nodes): 
        return ~hull.is_inside(nodes)
        
    def get_R(hull):
        return np.sqrt(np.nanmax(np.sum(np.square(hull.sample.G), axis=1)))
        
    def get_r(hull): 
        # zero at best, 
        # on the safe side, between -R and 0
        #r2_hull = np.nanmin(np.sum(np.square(hull.sample.G), axis=1))
        #return -np.sqrt(r2_hull)
        
        #č když na get_r nahlížím z hlediska toho,
        #č že r vymezuje mezikruží, kde nejsme jistí
        #č inside-outside,
        #č tak zde get_r vždy musí vracet prostě R
        return hull.get_R()
        
    def get_orth_outside(hull):
        x = np.full(hull.sample.nvar, hull.get_r())
        return calculate_brick_complement_probability(-x, x)
        
     
    def get_2FORM_outside(hull):
        return 2*stats.norm.sf(hull.get_r())
        
     
    def fire(hull, *args, **kwargs):
        pass
    
    def shot(hull, *args, **kwargs):
        pass
        


#č vyhejbám slovu Box, ať nevnáším ještě víc zmatku v označení
class BrickHull: #č nebo BoundingBrick
    def __init__(hull, sample, space='G'):
        hull.sample = sample
        hull.space = space
        
        
        hull._npoints = 0
        #č miny a maxy obsahují minima a maxima 
        #č podel jednotlivých souřadnic (proměnných)
        #č sincerely yours, Captain Obvious.
        hull.mins = np.full(hull.sample.nvar, np.inf)
        hull.maxs = np.full(hull.sample.nvar, -np.inf)
    
    def _update(hull):
        if hull._npoints < hull.npoints:
            hull.mins = np.nanmin(hull.points, axis=0)
            hull.maxs = np.nanmax(hull.points, axis=0)
            hull._npoints = hull.npoints
        
        
    def is_inside(hull, nodes):
        hull._update()
        x = getattr(nodes, hull.space)
        more = np.all(x < hull.maxs, axis=1)
        less = np.all(x > hull.mins, axis=1)
        
        return np.all((more, less), axis=0) #np.all(np.array((more, less)), axis=0)
        
    def is_outside(hull, nodes): 
        return ~hull.is_inside(nodes)
        
    def get_hyperplane_distances(hull, nodes):
        hull._update()
        x = getattr(nodes, hull.space)
        maxs = np.nanmax(x - hull.maxs, axis=1)
        mins = np.nanmax(hull.mins - x, axis=1)
        
        return np.max((maxs, mins), axis=0) #np.all(np.array((more, less)), axis=0)
        
    def get_design_points(hull):
        hull._update()
        #sample_model = -hull.A * hull.b.reshape(-1,1)
        sample_model = np.vstack((np.diag(hull.maxs), np.diag(hull.mins)))
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
        
    def get_exploration_vector(hull):
        hull._update()
        to_fire = np.nanargmax(hull.b)
        return hull.A[to_fire], hull.b[to_fire], hull.sample
        
    # shortcut for Ghull
    # valid only if space==G
    def get_r(hull):
        if hull.space=='G':
            hull._update()
            return np.min((-hull.mins, hull.maxs))
        else:
            return 0
        
    @property
    def points(hull):
        return getattr(hull.sample, hull.space)
        
    @property
    def npoints(hull):
        return len(hull.sample)
        
    @property
    def nsimplex(hull):
        return hull.sample.nvar*2
        
    @property
    def A(hull):
        hull._update()
        #č žádná optimizace, ale kdo tu funkci bude spouštět?
        diag = np.diag(np.ones(hull.sample.nvar))
        return np.vstack((diag, -diag))
        
    @property
    def b(hull):
        hull._update()
        return np.concatenate((hull.maxs, -hull.mins))
        
    @property
    def equations(hull):
        hull._update()
        #č žádná optimizace, ale kdo tu funkci bude spouštět?
        diag = np.diag(np.ones(hull.sample.nvar))
        
        A = np.vstack((diag, -diag))
        b = np.concatenate((-hull.maxs, hull.mins))
        return np.hstack((A,b[:,None]))
        
        
    def get_orth_outside(hull):
        if hull.space == 'G':
            hull._update()
            return calculate_brick_complement_probability(hull.mins, hull.maxs)
        else:
            #č když prostor není G, můžeme sice něco odvodit od s-ball 
            #č ale nechť se s tím pará GHull, vrátíme nulu.
            return 0
            
    def get_orth_equations(hull):
        return hull.equations
        
        
    def _2FORM_helper(hull):
        hull._update()
        #č je to úplně zbýtečně trapit se nejakejma argama
        #č poďme na to globálně!
        #č (doufám, že nvar*2 výpočtů cdf() nikomu neublíží... )
        #min_arg = np.nanargmin(hull.mins)
        #max_arg = np.nanargmax(hull.maxs)
        cdfs = stats.norm.cdf(hull.mins)
        sfs = stats.norm.sf(hull.maxs)
        #č _update() počítá nanmin a nanmax
        #č takže nemusíme se bat nějakého hnusu v číslech
        pfs = cdfs + sfs
        i = np.argmax(pfs) # number of variable
        # we'll return 2FORM estimation 
        # and corresponding number of variable
        return pfs[i], i
        
        
    def get_2FORM_outside(hull):
        if hull.space == 'G':
            # hull._update() will perform _FORM_helper()
            p_out, __i = hull._2FORM_helper()
            return p_out
        else:
            #č když prostor není G, můžeme sice něco odvodit od s-ball 
            #č ale nechť se s tím pará GHull, vrátíme nulu.
            return 0
            
        
    def get_2FORM_equations(hull):
        __p_out, i = hull._2FORM_helper()
        equations = np.zeros((2,hull.sample.nvar+1))
        equations[0, i] = 1
        equations[0, -1] = -hull.maxs[i]
        equations[1, i] = -1
        equations[1, -1] = hull.mins[i]
        return equations
        
    def shot(hull, *args, **kwargs):
        pass
        
        # add use_MC to the function signature
        # but remain unimplemented for now
        #č ten fire je hrozný. Výhodit U prostor 
        #č (jsou tam nepřesné cdf transformace)
        #č po úpravě přídat hull._update() na začátku.
    def fire(hull, ns, use_MC=False): # boom
        sample_U = hull.sample.U
        
        U_mins = np.nanmin(sample_U, axis=0)
        U_maxs = np.nanmax(sample_U, axis=0)
        
        #č min nebo max?
        if np.nanmax(U_mins) > (1-np.nanmin(U_maxs)):
            #č expandujeme se dolu
            var = np.nanargmax(U_mins)
            value = U_mins[var]
            t = np.linspace(value, 0, ns, endpoint=False)
        else:
            #č expandujeme se nahoru
            var = np.nanargmin(U_maxs)
            value = U_maxs[var]
            t = np.linspace(value, 1, ns, endpoint=False)
        
        boom = np.random.random((ns, hull.sample.nvar))
        boom[:, var] = t
        
        return hull.sample.f_model.new_sample(boom, space='U')
    

class DirectHull:
    # take some global functions
    fire = fire
    shot = shot
    _orth_helper = _orth_helper
    get_orth_outside = get_orth_outside
    get_orth_equations = get_orth_equations
    
    _2FORM_helper = _2FORM_helper
    get_2FORM_outside = get_2FORM_outside
    get_2FORM_equations = get_2FORM_equations
    
    
    def __init__(hull, sample, direct_plan, space='G'):
        hull.sample = sample
        hull.direct_plan = direct_plan
        hull.space = space
        
        hull.regen()
        
    def regen(hull):
        hull._npoints = 0
        
        hull.bp = np.full(len(hull.direct_plan), -np.inf)
        #hull.update()
    
    #č nejsem jist, jestli ten update vůbec dělat.
    #č lze navrhnout třidu aj tak, že sama bude hlídat změny.
    #č Jenomže, co kdybychom ten automatický update nechtěli?
    def _update(hull):
        if hull._npoints < hull.npoints:
            #hull.points = getattr(hull.sample, hull.space)
            new_points = hull.points[hull._npoints:]
            
            A = hull.direct_plan @ new_points.T
            new_bp = np.nanmax(A, axis=1)
            hull.bp = np.nanmax((new_bp, hull.bp), axis=0)
            
            hull._npoints = hull.npoints

    @property
    def points(hull):
        return getattr(hull.sample, hull.space)
        
    @property
    def npoints(hull):
        return len(hull.sample)
        
    @property
    def nsimplex(hull):
        return len(hull.direct_plan)
        
    @property
    def A(hull):
        return hull.direct_plan
        
    @property
    def b(hull):
        hull._update()
        return -hull.bp
        
    @property
    def equations(hull):
        return np.hstack((hull.A, hull.b[:,None]))

    def get_hyperplane_distances(self, nodes):
        self._update()
        x = getattr(nodes, self.space)
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        A = self.A
        b = self.b
        
        # N=ns, E - number of hyperplane equations
        ExN = A @ x.T + np.atleast_2d(b).T
        return np.nanmax(ExN, axis=0)
        
    def query(self, nodes):
        self._update()
        x = getattr(nodes, self.space)
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        A = self.A
        b = self.b
        
        # N=ns, E - number of hyperplane equations
        ExN = A @ x.T + np.atleast_2d(b).T
        i = np.argmax(ExN, axis=0)
        d = np.take_along_axis(ExN, np.atleast_2d(i), axis=0).reshape(-1)
        return d, i

    def is_inside(hull, nodes):
        hull._update()
        x = getattr(nodes, hull.space)
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        A = hull.direct_plan
        bp = np.atleast_2d(hull.bp).T
        
        # N=ns, E - number of hyperplane equations
        ExN = A @ x.T
        higher = np.all(ExN < bp, axis=0)
        return higher
    
    def is_outside(hull, nodes): 
        return ~hull.is_inside(nodes)
        
    def get_design_points(hull):
        hull._update()
        sample_model = -hull.A * hull.b.reshape(-1,1)
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
    
    def get_exploration_vector(hull):
        hull._update()
        to_fire = np.nanargmax(hull.b)
        return hull.A[to_fire], hull.b[to_fire], hull.sample
    
    # shortcut for Ghull
    # valid only if space==G
    def get_r(hull):
        if hull.space=='G':
            hull._update()
            return np.min(hull.bp)
        else:
            return 0
    
    


class CompleteHull:
    # take some global functions
    fire = fire
    shot = shot
    _orth_helper = _orth_helper
    get_orth_outside = get_orth_outside
    get_orth_equations = get_orth_equations
    
    _2FORM_helper = _2FORM_helper
    get_2FORM_outside = get_2FORM_outside
    get_2FORM_equations = get_2FORM_equations
    
    def __init__(hull, sample, direct_plan, space='G'):
        hull.sample = sample
        hull.direct_plan = direct_plan
        hull.space = space
        
        hull.regen()
        
    def regen(hull):
        hull._npoints = 0
        hull.mins = np.full(hull.sample.nvar, np.inf)
        hull.maxs = np.full(hull.sample.nvar, -np.inf)
        
        hull.bp = np.full(len(hull.direct_plan), -np.inf)
        hull.bm = np.full(len(hull.direct_plan), np.inf)
        
        #hull.update()
    
    #č nejsem jist, jestli ten update vůbec dělat.
    #č lze navrhnout třidu aj tak, že sama bude hlídat změny.
    #č Jenomže, co kdybychom ten automatický update nechtěli?
    def _update(hull):
        if hull._npoints < hull.npoints:
            #hull.points = getattr(hull.sample, hull.space)
            new_points = hull.points[hull._npoints:]
            
            new_mins = np.nanmin(new_points, axis=0)
            new_maxs = np.nanmax(new_points, axis=0)
            
            hull.mins = np.nanmin((new_mins, hull.mins), axis=0)
            hull.maxs = np.nanmax((new_maxs, hull.maxs), axis=0)
            
            A = hull.direct_plan @ new_points.T
            new_bp = np.nanmax(A, axis=1)
            new_bm = np.nanmin(A, axis=1)
            
            hull.bp = np.nanmax((new_bp, hull.bp), axis=0)
            hull.bm = np.nanmin((new_bm, hull.bm), axis=0)
            
            hull._npoints = hull.npoints

    @property
    def points(hull):
        return getattr(hull.sample, hull.space)
        
    @property
    def npoints(hull):
        return len(hull.sample)
        
    @property
    def nsimplex(hull):
        return 2 * (len(hull.direct_plan) + hull.sample.nvar)
        
    @property
    def A(hull):
        hull._update()
        #č žádná optimizace, ale kdo tu funkci bude spouštět?
        diag = np.diag(np.ones(hull.sample.nvar))
        return np.vstack((diag, -diag, -hull.direct_plan, hull.direct_plan))
        
    @property
    def b(hull):
        hull._update()
        return np.concatenate((-hull.maxs, hull.mins, hull.bm, -hull.bp))
        
    @property
    def equations(hull):
        hull._update()
        #č žádná optimizace, ale kdo tu funkci bude spouštět?
        diag = np.diag(np.ones(hull.sample.nvar))
        
        A = np.vstack((diag, -diag, -hull.direct_plan, hull.direct_plan))
        b = np.concatenate((-hull.maxs, hull.mins, hull.bm, -hull.bp))
        return np.hstack((A,b[:,None]))

    def is_inside(hull, nodes):
        hull._update()
        x = getattr(nodes, hull.space)
        
        more = np.all(x < hull.maxs, axis=1)
        less = np.all(x > hull.mins, axis=1)
        
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        A = hull.direct_plan
        bp = np.atleast_2d(hull.bp).T
        bm = np.atleast_2d(hull.bm).T
        
        # N=ns, E - number of hyperplane equations
        ExN = A @ x.T
        higher = np.all(ExN < bp, axis=0)
        lower = np.all(ExN > bm, axis=0)
        return np.all((more, less, higher, lower), axis=0)
    
    def is_outside(hull, nodes): 
        return ~hull.is_inside(nodes)
        
    def get_design_points(hull):
        hull._update()
        sample_model = -hull.A * hull.b.reshape(-1,1)
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
        
    # shortcut for Ghull
    # valid only if space==G
    def get_r(hull):
        if hull.space=='G':
            hull._update()
            return min(-np.max(hull.mins), np.min(hull.maxs), -np.max(hull.bm), np.min(hull.bp))
        else:
            return 0
    
            




QHullEstimation = namedtuple('QHullEstimation', (
                                                'nsim',
                                                'nvar',
                                                'nfacets',
                                                'r', 'R',
                                                'inner',
                                                'shell',
                                                'outer',
                                                'chi_outside',
                                                'orth_outside',
                                                #'tn_scheme',
                                                #'tn_scheme_points',
                                                'inside',
                                                'outside'
                                                ))


def get_nsphere_const(nvar, const=1):
    if nvar == 2:
        return const * 2 * np.pi
    elif nvar == 1:
        return const * 2
    else:
        return get_nsphere_const(nvar - 2, const * 2 * np.pi / (nvar - 2) )

class QHull:
    # take some global function
    _orth_helper = _orth_helper
    #get_orth_outside = get_orth_outside
    get_orth_equations = get_orth_equations
    
    _2FORM_helper = _2FORM_helper
    #get_2FORM_outside = get_2FORM_outside
    get_2FORM_equations = get_2FORM_equations
    
    def __init__(self, sample, space='G', incremental=True, auto_update=True, tn_scheme=None):
        self.sample = sample
        self.incremental = incremental
        self.space = space
        self.auto_update = auto_update
        
        self.sball = sball.Sball(sample.nvar)
        self.shell = sball.Shell(sample.nvar)
        self.fallback_plan = quadpy.un.mysovskikh_1(sample.nvar).points
        
        if tn_scheme is None:
            self.tn_scheme = quadpy.tn.grundmann_moeller(sample.nvar - 1, 5)
        else:
            self.tn_scheme = tn_scheme
        
        nvar = sample.nvar
        #self.chi_const = 2**((nvar - 1) / 2) * special.gamma(nvar - 1/2) / special.gamma(nvar / 2) 
        self.chi_like_const = np.sqrt(np.pi) / np.sqrt(2) / (2 * np.pi)**(nvar/2)
        self.nsphere_surface_area = get_nsphere_const(nvar)
    
    
    
    def regen(self):
        points = getattr(self.sample, self.space)
        self.convex_hull = spatial.ConvexHull(points, incremental=self.incremental)
        
    
    def __getattr__(self, attr):
        #č branime se rekurzii
        # defend against recursion
        #оӵ рекурсилы пезьдэт!
        if attr == 'convex_hull':
            #č zkusme rychle konvexní obálky sestavit
            #č a ihned ji vrátit
            self.regen()
            return self.convex_hull
        
        elif attr == 'enough_points':
            return self.sample.nvar < self.sample.nsim
            
        elif attr == 'A':
            return self.convex_hull.equations[:,:-1]
        elif attr == 'b':
            return self.convex_hull.equations[:,-1]
            
        elif attr == 'points':
            if self.enough_points:
                return self.convex_hull.points
            else:
                return getattr(self.sample, self.space)
                
        elif attr == 'npoints':
            if self.enough_points:
                return self.convex_hull.npoints
            else:
                return len(self.sample)
        
        elif attr == 'nsimplex':
            if self.enough_points:
                return self.convex_hull.nsimplex
            else:
                return 0
            
            
        #ё По всем вопросам обращайтесь 
        #ё на нашу горячую линию    
        else:
            self._update() #č dycky čerstý chleba!
            return getattr(self.convex_hull, attr)
            
    def _update(hull):
        if hull.auto_update:
            hull.update()            
    
    def update(self):
        if self.enough_points and (self.convex_hull.npoints < self.sample.nsim):
            if self.incremental:
                points = getattr(self.sample, self.space)  
                self.convex_hull.add_points(points[self.convex_hull.npoints:])
            else:
                self.regen()
    
    
    def get_hyperplane_distances(self, nodes):
        if self.enough_points:
            self._update()
            x = getattr(nodes, self.space)
            
            #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
            A = self.convex_hull.equations[:,:-1]
            b = self.convex_hull.equations[:,-1]
            
            # N=ns, E - number of hyperplane equations
            ExN = A @ x.T + np.atleast_2d(b).T
            return np.nanmax(ExN, axis=0)
        else:
            return DirectHull(self.sample, self.fallback_plan, self.space).get_hyperplane_distances(nodes)
        
    def query(self, nodes):
        if self.enough_points:
            self._update()
            x = getattr(nodes, self.space)
            
            #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
            A = self.convex_hull.equations[:,:-1]
            b = self.convex_hull.equations[:,-1]
            
            # N=ns, E - number of hyperplane equations
            ExN = A @ x.T + np.atleast_2d(b).T
            i = np.argmax(ExN, axis=0)
            d = np.take_along_axis(ExN, np.atleast_2d(i), axis=0).reshape(-1)
            return d, i
        else:
            return DirectHull(self.sample, self.fallback_plan, self.space).query(nodes)
        
    def is_inside(self, nodes):
        if self.enough_points:
            self._update()
            x = getattr(nodes, self.space)
            
            #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
            A = self.convex_hull.equations[:,:-1]
            b = self.convex_hull.equations[:,-1]
            
            # N=ns, E - number of hyperplane equations
            ExN = A @ x.T + np.atleast_2d(b).T
            mask = np.all(ExN < 0, axis=0)
            return mask
        else:
            return np.full(len(nodes), False)
    
    def is_outside(hull, nodes): 
        if hull.enough_points:
            return ~hull.is_inside(nodes)
        else:
            return np.full(len(nodes), True)
        
    def get_design_points(hull):
        hull._update()
        sample_model = -hull.A * hull.b.reshape(-1,1)
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
        
    def get_design_point(hull):
        hull._update()
        to_fire = np.nanargmax(hull.b) #č tak, pro jistotu
        sample_model = -hull.A[to_fire] * hull.b[to_fire]
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
    
    def get_exploration_vector(hull):
        hull._update()
        if hull.enough_points:
            to_sample = np.nanargmax(hull.b)
            sample = hull.sample[hull.simplices[to_sample]]
            return hull.A[to_sample], hull.b[to_sample], sample
        else:
            return DirectHull(hull.sample, hull.fallback_plan, hull.space).get_exploration_vector()
        
        
    # shortcut for Ghull
    # valid only if space==G
    def get_r(hull):
        if hull.space=='G':
            if hull.enough_points:
                hull._update()
                b = hull.convex_hull.equations[:,-1]
                return -np.nanmax(b)
            else:
                return 0
        else:
            return 0
    
    def get_R(self):
        assert self.space == 'G'
        sum_squared = np.sum(np.square(self.sample.G), axis=1)
        #index = np.argmax(sum_squared)
        return np.sqrt(np.nanmax(sum_squared))
    
    def shot(hull, ns, use_MC=False):
        try:
            # take global function for shot()
            return shot(hull, ns, use_MC)
        except:
            pass
            
    def fire(hull, ns, use_MC=False):
        try:
            # take global function for fire()
            return fire(hull, ns, use_MC)
        except:
            pass
    
    def get_orth_outside(hull):
        try:
            return get_orth_outside(hull)
        except:
            #č asi máme problém s triangulací 
            #č odvažně vratíme jedničku
            #č neřikejte mi, že musím pěčlivějc zpracovavat chyby!
            return 1
    
    def get_2FORM_outside(hull):
        try:
            return get_2FORM_outside(hull)
        except:
            #č asi máme problém s triangulací 
            #č odvažně vratíme jedničku
            #č neřikejte mi, že musím pěčlivějc zpracovavat chyby!
            return 1
    
    def get_norm_outside(hull):
        hull._update()
        if hull.space != 'G':
            raise
        
        vertices_model = hull.points[hull.simplices].transpose((1, 0, 2))
        result = hull.tn_scheme.integrate(hull._outside_callback, vertices_model)
        
        return np.sum(result) * hull.chi_like_const
        
        
    def get_chi_outside(hull):
        hull._update()
        if hull.space != 'G':
            raise
        
        
        vertices_model = hull.points[hull.simplices].transpose((1, 0, 2))
        result = hull.tn_scheme.integrate(hull._chi_outside_callback, vertices_model)
        mask = result > 0 #č takto zapsana podmínka pokryvá i Nan
        if not np.all(mask):
            number = len(mask) - np.count_nonzero(mask)
            print("QHull: Negative values in %s facets have appeared during integration" % number)
            return np.sum(result, where=mask) / hull.nsphere_surface_area
        return np.sum(result) / hull.nsphere_surface_area
        
#    # quadpy
#    def _norm_inside_callback(hull, x):
#        # x.shape == (simplex_dim + 1, nsimplex, scheme_npoints) #(3, 26, 56)
#        
#        cdfs = stats.norm.cdf(x)
#        pdfs = stats.norm.pdf(x)
#        
#        for i in range(hull.sample.nvar):
#            cdfs[i] *= np.prod(pdfs[:i], axis=0)
#            cdfs[i] *= np.prod(pdfs[i+1:], axis=0)
#            cdfs[i] *= np.atleast_2d(hull.A[:, i]).T 
#        
#        
#        # n_values x nsimplex x scheme_points
#        return cdfs

        
    
    
    def _outside_callback(hull, x):
        # x.shape == (simplex_dim + 1, nsimplex, scheme_npoints) #(3, 26, 56)
        
        #č přijde velké, obrovské numpy pole
        #č a já bych chtěl, aby se tu zbytečně nealokovalo
        
        result = np.einsum('ijk,ijk->jk', x, x) # r**2
        np.sqrt(result, out=result) # r
        
        #result = r / np.sqrt(2) # r / sqrt(2)
        np.divide(result, np.sqrt(2), out=result)
        #special.gammaincc(hull.sample.nvar - 1/2, result, out=result) # F
        special.erfc(result, out=result) 
        
#        #č normalizujeme normály. Jenže normály samotné dodáme později
#        result /= r 
#        
#        # A = nsimplex x nvar
#        # x = nvar x nsimplex x scheme_npoints
#        # normals = nsimplex x scheme_npoints
#        #normals = np.einsum('ij,jik->ik', hull.A, x)
#        np.einsum('ij,jik->ik', hull.A, x, out=r) # normals
#        result *= r 
        
        # n_values x nsimplex x scheme_points
        return result 

    def _outside_normals_callback(hull, x):
        # x.shape == (simplex_dim + 1, nsimplex, scheme_npoints) #(3, 26, 56)
        
        #č přijde velké, obrovské numpy pole
        #č a já bych chtěl, aby se tu zbytečně nealokovalo
        
        r = np.einsum('ijk,ijk->jk', x, x) # r**2
        np.sqrt(r, out=r) # r
        
        result = r / np.sqrt(2) # r / sqrt(2)
        #np.divide(result, np.sqrt(2), out=result)
        #special.gammaincc(hull.sample.nvar - 1/2, result, out=result) # F
        special.erfc(result, out=result) 
        
        #č normalizujeme normály. Jenže normály samotné dodáme později
        result /= r 
        
        # A = nsimplex x nvar
        # x = nvar x nsimplex x scheme_npoints
        # normals = nsimplex x scheme_npoints
        #normals = np.einsum('ij,jik->ik', hull.A, x)
        np.einsum('ij,jik->ik', hull.A, x, out=r) # normals
        result *= r 
        
        # n_values x nsimplex x scheme_points
        return result 


    def _chi_outside_callback(hull, x):
        # x.shape == (simplex_dim + 1, nsimplex, scheme_npoints) #(3, 26, 56)
        
        nvar = hull.sample.nvar
        
        #č přijde velké, obrovské numpy pole
        #č a já bych chtěl, aby se tu zbytečně nealokovalo
        
        result = np.einsum('ijk,ijk->jk', x, x) # r**2
        r_nvar = np.power(result, nvar/2) # r**(nvar-1) (chi) * r (to normalize)
        
        np.divide(result, 2, out=result) # r**2 / 2
        special.gammaincc(nvar / 2, result, out=result) # F
        
        #č normalizujeme normály. Jenže normály samotné dodáme později
        result /= r_nvar 
        
        # A = nsimplex x nvar
        # x = nvar x nsimplex x scheme_npoints
        # normals = nsimplex x scheme_npoints
        #normals = np.einsum('ij,jik->ik', hull.A, x)
        np.einsum('ij,jik->ik', hull.A, x, out=r_nvar) # normals
        result *= r_nvar
        
        # n_values x nsimplex x scheme_points
        return result 

    def get_convex_hull_estimation(hull):
        hull._update()
        
        r = hull.get_r()
        R = hull.get_R()
        if r<0:
            hull.shell.set_bounds(0, R)
        else:
            hull.shell.set_bounds(r, R)
        
        outer = hull.shell.pf
        shell = hull.shell.p_shell
        
        if hull.enough_points:
            chi_outside = hull.get_chi_outside()
        else:
            chi_outside = 0
        
        orth_outside = hull.get_orth_outside()
        
        r_outside = outer + shell
        outside = min(r_outside, chi_outside)
        outside = max(outside, orth_outside, outer)
        
        return QHullEstimation(
                              hull.npoints,
                              hull.sample.nvar,
                              hull.nsimplex,
                              r, R,
                              hull.shell.ps,
                              shell,
                              outer,
                              chi_outside,
                              orth_outside,
                              #'tn_scheme',
                              #'tn_scheme_points',
                              1 - outside,
                              outside
                              )
                        

class QHullCubature(QHull):
    
    def get_chi_outside(hull):
        hull._update()
        
        if hull.space != 'G':
            raise
        
        
        integrals = 0
        # if first time runned
        if 'facets' not in dir(hull):
            hull.facets = facets = dict()
            ncubature = len(hull.tn_scheme.weights)
            hull._x = np.empty((hull.sample.nvar, ncubature))
            hull._r = np.empty(ncubature)
            hull._result = np.empty(ncubature)
            for indices, normal in zip(hull.simplices, hull.A):
                facet_integral = hull.get_facet_outside(indices, normal)
                facets[indices.tobytes()] = facet_integral
                integrals += facet_integral
            
            return integrals / hull.nsphere_surface_area
        
        facets = hull.facets
        new_facet_dict = dict()
        for indices, normal in zip(hull.simplices, hull.A):
            key = indices.tobytes()
            if key in facets:
                facet_integral = facets[key]
            else:
                facet_integral = hull.get_facet_outside(indices, normal)
            new_facet_dict[key] = facet_integral
            integrals += facet_integral
            
        hull.facets = new_facet_dict
        return integrals / hull.nsphere_surface_area
            

            

    def get_facet_outside(hull, indices, normal):
        vertices_model = hull.points[indices]
        
        vol = quadpy.tn.get_vol(vertices_model)
        if not np.isfinite(vol):
            print("QHull: Incorrect area has occured in facet %s" % indices)
            return 0
        
        nvar = len(normal)
        x = hull._x
        r_nvar = hull._r
        result = hull._result
        
        # nvar x n_integration_nodes
        # Transform the points `xi` from the reference simplex onto `simplex`.
        # same as quadpy.tn.transform(hull.tn_scheme.points, vertices_model.T)
        #np.dot(simplex, points)
        np.matmul(vertices_model.T, hull.tn_scheme.points, out=x)
        
        np.einsum('ij,ij->j', x, x, out=result) # r**2
        np.power(result, nvar/2, out=r_nvar) # r**(nvar-1) (chi) * r (to normalize)
        
        np.divide(result, 2, out=result) # r**2 / 2
        special.gammaincc(nvar / 2, result, out=result) # F
        
        #č normalizujeme normály. Jenže normály samotné dodáme později
        result /= r_nvar 
        
        #č normála
        np.dot(normal, x, out=r_nvar)
        result *= r_nvar
        
        # facet_outside = integral / unit_ball_surface_area
        integral = vol * np.inner(result, hull.tn_scheme.weights)
        if integral > 0:
            return integral
        else:
            print("QHull: Negative value has appeared in facet %s during integration" % indices)
            return 0



# Gaussian brick
#č Gaussovská cihla
#
# it was a mistake to design previous classes without explicit update()
class Grick:
    # take some global functions
    fire = fire
    shot = shot
    
    
    def get_FORM_outside(self):
        return stats.norm.sf(self.get_r())
    
    
    def __init__(hull, sample, direct_plan, nrandom=50, auto_update=True):
        hull.sample = sample
        hull.direct_plan = direct_plan
        hull.space = 'G'
        hull.ndim = sample.nvar
        hull.nrandom = nrandom #č počet náhodných normalových věktorů
        
        hull.auto_update = auto_update
        
        hull.regen()
        
        
    def regen(hull):
        hull._npoints = 0
        hull.mins = np.full(hull.ndim, np.inf)
        hull.maxs = np.full(hull.ndim, -np.inf)
        
        hull.bp_dp = np.full(len(hull.direct_plan), -np.inf)
        hull.bm_dp = np.full(len(hull.direct_plan), np.inf)
        
        hull.bp = np.full(hull.ndim, -np.inf)
        hull.bm = np.full(hull.ndim, np.inf)
        
        hull.orth_basis = np.eye(hull.ndim) 
        
        hull.update()
    
    
    def _update(hull):
        if hull.auto_update:
            hull.update()
    
    # it was a mistake to design previous classes without explicit update()
    def update(hull):
        if hull._npoints < hull.npoints:
            hull._points = hull.sample.G
            
            new_points = hull._points[hull._npoints:]
            
            new_mins = np.nanmin(new_points, axis=0)
            new_maxs = np.nanmax(new_points, axis=0)
            
            np.nanmin((new_mins, hull.mins), axis=0, out=hull.mins)
            np.nanmax((new_maxs, hull.maxs), axis=0, out=hull.maxs)
            
            new_bp, new_bm = get_bp_bm(hull.direct_plan, new_points)
            np.nanmax((new_bp, hull.bp_dp), axis=0, out=hull.bp_dp)
            np.nanmin((new_bm, hull.bm_dp), axis=0, out=hull.bm_dp)
            
            
            new_bp, new_bm = get_bp_bm(hull.orth_basis, new_points)
            np.nanmax((new_bp, hull.bp), axis=0, out=hull.bp)
            np.nanmin((new_bm, hull.bm), axis=0, out=hull.bm)
            
            hull._gen_orth_basis()
                
            hull._npoints = hull.npoints


    def gen_random_Ab(hull):
        ns = hull.ndim * 2 + len(hull.direct_plan) + hull.nrandom
        #č necháme vygenerovat maticu náhodných čisel
        #č ale použijeme jen část z ní
        #č moje zkušenost je, že np.ones a np.random.randn běží stejně rychle
        A = np.random.randn(ns, hull.ndim) #random directions
        A[:hull.ndim] = np.eye(hull.ndim)
        A[hull.ndim:hull.ndim*2] = hull.orth_basis
        A[hull.ndim*2:ns-hull.nrandom] = hull.direct_plan
        
        # rand_dir: prepare ns random directions on a unit d-sphere
        rand_dir = A[ns-hull.nrandom:]
        
        lengths = np.sum(np.square(rand_dir), axis=1)
        lengths = np.sqrt(lengths, out=lengths) #lengths of each radius-vector
        
        # scale all radii-vectors to unit length
        # use [:,None] to get an transposed 2D array
        np.divide(rand_dir, lengths[:,None], out=rand_dir) 
        
        b = np.empty(ns) 
        #č Nám jde o basis. Dáme ty nejlepší b-čka.
        #č znaménka vektorů, tj. směry dopředu nebo dozadu v této fázi nevadí
        np.nanmax((-hull.maxs, hull.mins), axis=0, out=b[:hull.ndim])
        np.nanmax((-hull.bp, hull.bm), axis=0, out=b[hull.ndim:hull.ndim*2])
        np.nanmax((-hull.bp_dp, hull.bm_dp), axis=0, out=b[hull.ndim*2:ns-hull.nrandom])
        
        rbp, rbm = get_bp_bm(rand_dir, hull._points)
        np.nanmax((-rbp, rbm), axis=0, out=b[ns-hull.nrandom:])
        
        return A, b


    def _gen_orth_basis(hull):
        A, b = hull.gen_random_Ab()
        
        #č je to špatné z hlediska testování,
        #č ale nevím jak z toho.
        #č pro 2D musím udělat vyjimku z "předposlední"
        if hull.ndim > 2:
            hull.orth_basis = _get_nD_orth_basis_from_normals(A, b)
        else:
            to_fire = np.nanargmax(b)
            a = A[to_fire]
            x, y = a
            hull.orth_basis = np.array([[x, y],[-y, x]])
            
        
        #č musí tam bejt G coordinates
        hull.bp, hull.bm = get_bp_bm(hull.orth_basis, hull._points)
        
        
        
        
    
    
        
        
    def get_orth_outside(hull):
        hull._update()
        return calculate_brick_complement_probability(hull.bm, hull.bp)


    def get_orth_equations(hull):
        hull._update()
        return hull.equations


    def get_2FORM_outside(hull):
        hull._update()
        return stats.norm.cdf(hull.bm[0]) + stats.norm.sf(hull.bp[0])
                
    def get_2FORM_equations(hull):
        hull._update()
        A = np.vstack((-hull.orth_basis[0], hull.orth_basis[0]))
        b = np.array([[hull.bm[0]],[-hull.bp[0]]])
        return np.hstack((A,b))  


    @property
    def points(hull):
        if hull.auto_update:
            return getattr(hull.sample, hull.space)
        else:
            return hull._points
    
        
    @property
    def npoints(hull):
        if hull.auto_update:
            return len(hull.sample)
        else:
            return hull._npoints
        
    
    
    @property #č stejně nikdo to nebude spouštět
    def nsimplex(hull):
        return hull.sample.nvar * 2
        
        
    @property
    def A(hull):
        hull._update()
        #č žádná optimizace, ale stejně nikdo to nebude spouštět
        return np.vstack((-hull.orth_basis, hull.orth_basis))
        
    @property
    def b(hull):
        hull._update()
        return np.concatenate((hull.bm, -hull.bp))
        
    @property
    def equations(hull):
        hull._update()
        #č žádná optimizace, ale stejně nikdo to nebude spouštět
        A = np.vstack((-hull.orth_basis, hull.orth_basis))
        b = np.concatenate((hull.bm, -hull.bp))
        return np.hstack((A,b[:,None]))

    def is_inside(hull, nodes):
        hull._update()
        x = nodes.G
        
        #E [normal, offset] forming the hyperplane equation of the facet (see Qhull documentation for more)
        bp = np.atleast_2d(hull.bp).T
        bm = np.atleast_2d(hull.bm).T
        
        # N=ns, E - number of hyperplane equations
        ExN = hull.orth_basis @ x.T
        higher = np.all(ExN < bp, axis=0)
        lower = np.all(ExN > bm, axis=0)
        return np.all((higher, lower), axis=0)
    
    def is_outside(hull, nodes): 
        return ~hull.is_inside(nodes)
        
    def get_design_points(hull):
        hull._update()
        sample_model = -hull.A * hull.b.reshape(-1,1)
        return hull.sample.f_model.new_sample(sample_model, space=hull.space)
        
    def get_exploration_vector(hull):
        hull._update()
        if hull.bm[0] > (-hull.bp[0]):
            return -hull.orth_basis[0], hull.bm[0], hull.sample
        else:
            return hull.orth_basis[0], -hull.bp[0], hull.sample
        
        
        
    def get_r(hull):
        hull._update()
        "calculates minimal signed distance to planes. Can therefore be negative"
        #č ukazalo se, že náhodou další orthogonály můžou trefit menší poloměr
        #return min(-hull.bm[0], hull.bp[0])
        return np.min((-hull.bm, hull.bp))
            
            
            
            
            
            

        





#!/usr/bin/env python
# coding: utf-8

"""
We will prepare here WhiteBox instances
of different well-known different dimension problems
"""

import numpy as np
from .. import g_models as gm
from .. import f_models
from ..whitebox import WhiteBox
import scipy.stats as stats
import mpmath # for suspension_3D

__all__ = [
            'nonlinear_oscilator_6D',
            'single_spring_nonlinear_oscilator_5D',
            #'noisy_lsf_Papaioannou_6D',
            #'noisy_lsf_Liu_6D',
            'passive_vehicle_suspension_3D'
            ]



def nonlinear_oscilator_6D():
    """
    DOI:10.1016/j.strusafe.2004.11.001
    Beneﬁt of splines and neural networks in simulation based
    structural reliability analysis
    """
    f = f_models.Norm(np.array([1, 1, 0.1, 0.5, 1, 1]), 
                        np.array([0.05, 0.10, 0.01, 0.05, 0.20, 0.20]))
    wt = WhiteBox(f, gm.nonlinear_oscilator)
    wt.pf_exact = 0.0281
    wt.pf_exact_method = 'known value'
    return wt

def single_spring_nonlinear_oscilator_5D():
    c_mean = 1 + 0.1
    c_std = np.sqrt(0.10**2 + 0.01**2)
    f = f_models.Norm(np.array([1, c_mean, 0.5, 1, 1]), 
                        np.array([0.05, c_std , 0.05, 0.20, 0.20]))
    wt = WhiteBox(f, gm.single_spring_nonlinear_oscilator)
    wt.pf_exact = 0.0281
    wt.pf_exact_method = 'known value'
    return wt


__all__.append("nonlinear_oscilator_pf6_6D")
def nonlinear_oscilator_pf6_6D():
    """
    @article{ECHARD2013232,
title = {A combined Importance Sampling and Kriging reliability method for small failure probabilities with time-demanding numerical models},
journal = {Reliability Engineering & System Safety},
volume = {111},
pages = {232-240},
year = {2013},
issn = {0951-8320},
doi = {https://doi.org/10.1016/j.ress.2012.10.008},
url = {https://www.sciencedirect.com/science/article/pii/S0951832012002086},
author = {B. Echard and N. Gayton and M. Lemaire and N. Relun},
keywords = {Reliability, Kriging metamodel, Surrogate model, Small failure probability, Importance sampling},
abstract = {Applying reliability methods to a complex structure is often delicate for two main reasons. First, such a structure is fortunately designed with codified rules leading to a large safety margin which means that failure is a small probability event. Such a probability level is difficult to assess efficiently. Second, the structure mechanical behaviour is modelled numerically in an attempt to reproduce the real response and numerical model tends to be more and more time-demanding as its complexity is increased to improve accuracy and to consider particular mechanical behaviour. As a consequence, performing a large number of model computations cannot be considered in order to assess the failure probability. To overcome these issues, this paper proposes an original and easily implementable method called AK-IS for active learning and Kriging-based Importance Sampling. This new method is based on the AK-MCS algorithm previously published by Echard et al. [AK-MCS: an active learning reliability method combining Kriging and Monte Carlo simulation. Structural Safety 2011;33(2):145–54]. It associates the Kriging metamodel and its advantageous stochastic property with the Importance Sampling method to assess small failure probabilities. It enables the correction or validation of the FORM approximation with only a very few mechanical model computations. The efficiency of the method is, first, proved on two academic applications. It is then conducted for assessing the reliability of a challenging aerospace case study submitted to fatigue.}
}
    """
    
    T1_mean = 1
    T1_std = 0.2 * T1_mean
    F1_mean = 0.6
    F1_std = F1_mean / 6
    #m, c1, c2, r, F1, t1 = sample.T
    f = f_models.Norm(np.array([1, 1, 0.1, 0.5, F1_mean, T1_mean]), 
                        np.array([0.05, 0.10, 0.01, 0.05, F1_std, T1_std]))
    wt = WhiteBox(f, gm.nonlinear_oscilator)
    wt.pf_exact = 9.09e-6
    wt.pf_exact_method = 'known value'
    return wt


__all__.append("single_spring_nonlinear_oscilator_pf6_5D")
def single_spring_nonlinear_oscilator_pf6_5D():
    c_mean = 1 + 0.1
    c_std = np.sqrt(0.10**2 + 0.01**2)
    T1_mean = 1
    T1_std = 0.2 * T1_mean
    F1_mean = 0.6
    F1_std = F1_mean / 6
    f = f_models.Norm(np.array([1, c_mean, 0.5, F1_mean, T1_mean]), 
                        np.array([0.05, c_std , 0.05, F1_std, T1_std]))
    wt = WhiteBox(f, gm.single_spring_nonlinear_oscilator)
    wt.pf_exact = 9.09e-6
    wt.pf_exact_method = 'known value'
    return wt






__all__.append("nonlinear_oscilator_pf8_6D")
def nonlinear_oscilator_pf8_6D():
    """
    @article{ECHARD2013232,
title = {A combined Importance Sampling and Kriging reliability method for small failure probabilities with time-demanding numerical models},
journal = {Reliability Engineering & System Safety},
volume = {111},
pages = {232-240},
year = {2013},
issn = {0951-8320},
doi = {https://doi.org/10.1016/j.ress.2012.10.008},
url = {https://www.sciencedirect.com/science/article/pii/S0951832012002086},
author = {B. Echard and N. Gayton and M. Lemaire and N. Relun},
keywords = {Reliability, Kriging metamodel, Surrogate model, Small failure probability, Importance sampling},
abstract = {Applying reliability methods to a complex structure is often delicate for two main reasons. First, such a structure is fortunately designed with codified rules leading to a large safety margin which means that failure is a small probability event. Such a probability level is difficult to assess efficiently. Second, the structure mechanical behaviour is modelled numerically in an attempt to reproduce the real response and numerical model tends to be more and more time-demanding as its complexity is increased to improve accuracy and to consider particular mechanical behaviour. As a consequence, performing a large number of model computations cannot be considered in order to assess the failure probability. To overcome these issues, this paper proposes an original and easily implementable method called AK-IS for active learning and Kriging-based Importance Sampling. This new method is based on the AK-MCS algorithm previously published by Echard et al. [AK-MCS: an active learning reliability method combining Kriging and Monte Carlo simulation. Structural Safety 2011;33(2):145–54]. It associates the Kriging metamodel and its advantageous stochastic property with the Importance Sampling method to assess small failure probabilities. It enables the correction or validation of the FORM approximation with only a very few mechanical model computations. The efficiency of the method is, first, proved on two academic applications. It is then conducted for assessing the reliability of a challenging aerospace case study submitted to fatigue.}
}
    """
    
    T1_mean = 1
    T1_std = 0.2 * T1_mean
    F1_mean = 0.45
    F1_std = F1_mean / 6
    #m, c1, c2, r, F1, t1 = sample.T
    f = f_models.Norm(np.array([1, 1, 0.1, 0.5, F1_mean, T1_mean]), 
                        np.array([0.05, 0.10, 0.01, 0.05, F1_std, T1_std]))
    wt = WhiteBox(f, gm.nonlinear_oscilator)
    wt.pf_exact = 1.55e-8
    wt.pf_exact_method = 'known value'
    return wt


__all__.append("single_spring_nonlinear_oscilator_pf8_5D")
def single_spring_nonlinear_oscilator_pf8_5D():
    c_mean = 1 + 0.1
    c_std = np.sqrt(0.10**2 + 0.01**2)
    T1_mean = 1
    T1_std = 0.2 * T1_mean
    F1_mean = 0.45
    F1_std = F1_mean / 6
    f = f_models.Norm(np.array([1, c_mean, 0.5, F1_mean, T1_mean]), 
                        np.array([0.05, c_std , 0.05, F1_std, T1_std]))
    wt = WhiteBox(f, gm.single_spring_nonlinear_oscilator)
    wt.pf_exact = 1.55e-8
    wt.pf_exact_method = 'known value'
    return wt




def noisy_lsf_Papaioannou_6D():
    """
    Papaioannou. Sequental importance sampling
    for structural reliability analysis
    """
    f = f_models.UnCorD([ #č já jsem nepřišel na to, jak to mysleli s tím rozdělením
                        stats.lognorm(s=8, loc=120), 
                        stats.lognorm(s=8, loc=120), 
                        stats.lognorm(s=8, loc=120), 
                        stats.lognorm(s=8, loc=120), 
                        stats.lognorm(s=10, loc=50), 
                        stats.lognorm(s=8, loc=40), 
                        ])
    wt = WhiteBox(f, gm.noisy_lsf)
    wt.pf_exact = 5.29e-4
    wt.pf_exact_method = 'known value'
    return wt




def noisy_lsf_Liu_6D():
    """
    @article{LIU1991161,
    title = {Optimization algorithms for structural reliability},
    journal = {Structural Safety},
    volume = {9},
    number = {3},
    pages = {161-177},
    year = {1991},
    issn = {0167-4730},
    doi = {https://doi.org/10.1016/0167-4730(91)90041-7},
    url = {https://www.sciencedirect.com/science/article/pii/0167473091900417},
    author = {Pei-Ling Liu and Armen {Der Kiureghian}},
    keywords = {algorithms, finite element, optimization, probability theory, reliability analysis, robustness, safety, structures},
    abstract = {Several optimization algorithms are evaluated for application in structural reliability, where the minimum distance from the origin to the limit-state surface in the standard normal space is required. The objective is to determine the suitability of the algorithms for application to linear and nonlinear finite element reliability problems. After a brief review, five methods are compared through four numerical examples. Comparison criteria are the generality, robustness, efficiency, and capacity of each method.}
    }
    """
    f = f_models.UnCorD([ #č já jsem nepřišel na to, jak to mysleli s tím rozdělením
                        stats.lognorm(s=1, loc=120, scale=12), 
                        stats.lognorm(s=1, loc=120, scale=12), 
                        stats.lognorm(s=1, loc=120, scale=12), 
                        stats.lognorm(s=1, loc=120, scale=12), 
                        stats.lognorm(s=1, loc=50, scale=15), 
                        stats.lognorm(s=1, loc=40, scale=12), 
                        ])
    wt = WhiteBox(f, gm.noisy_lsf)
    wt.pf_exact = stats.norm.sf(2.3482)
    wt.DP = [117.3, 115.3, 115.3, 117.3, 83.62, 55.54]
    wt.pf_exact_method = 'known value'
    return wt






def passive_vehicle_suspension_3D():
    mu1, mu2, mu3 = (431.7221, 1475.5503, 55.0406)
    s = 10
    f3 = f_models.Norm(np.array((mu1, mu2, mu3)),
                        np.array((s, s, s))
                        )
    wt = WhiteBox(f3, gm.PassiveVehicleSuspension())
    # first (bottom) design point 
    max_failure_1 = 1442.64218624
    min_success_1 = 1442.64243976
    DP1 = 1442.642281
    # pf(beta 1) je mezi array([0.00049949, 0.00049954])

    # second (left) design point 
    max_failure_2 = 391.20793653
    min_success_2 = 391.21532796
    DP2 = 391.2123922
    # pf(beta 2) je mezí [2.54542636e-05, 2.55348088e-05]

    # ještě 3 poruchových bodů on very low K values, however, they do not form simple failure plane, there are success points even lower.
    #array([[4.08598926e+02, 1.47388406e+03, 1.00615452e-01],
    #   [4.77325547e+02, 1.46414564e+03, 5.55312946e-02],
    #   [4.24694913e+02, 1.45236644e+03, 9.50311636e-01]])
    
    wt.pf_upper = 1 - (1 - mpmath.ncdf(min_success_2, mu=mu1, sigma=s)) \
                * (1 - mpmath.ncdf(min_success_1, mu=mu2, sigma=s)) \
                * (1 - mpmath.ncdf("9.50311636e-01", mu=mu3, sigma=s))
                
    wt.pf_lower = 1 - (1 - mpmath.ncdf(max_failure_2, mu=mu1, sigma=s)) \
                * (1 - mpmath.ncdf(max_failure_1, mu=mu2, sigma=s))
    wt.pf_exact = float(1 - (1 - mpmath.ncdf(DP2, mu=mu1, sigma=s)) \
                * (1 - mpmath.ncdf(DP1, mu=mu2, sigma=s)))
    #č 0.000525 je hodně dobrý odhad
    wt.pf_exact_method = 'FORM-based precise geometry analysis'
    
    wt.r_exact = (mu2 - DP1) / s
    
    #wt.pf_exact = 0.00052
    #wt.pf_exact_method = 'IS'
    #wt.Nsim = 1000000
    #wt.description = ""
    return wt





import sys
import numpy as np
from .. import schemes
from ..candybox import CandyBox
#from ..dicebox.goal import Goal, Razitko, DiceBox
from ..dicebox.circumtri import CirQTri, CircumTri, QTri
from ..dicebox._exploration import DumbExploration
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtWidgets

spaces = ['R', 'aR', 'Rn', 'aRn', 'P', 'GK', 'G', 'aG', 'U', 'aU']
potentials = ['q_psee', 'psee', 'fee', 'fee2', 'ksee', 'chee', 'chee2', 'dd']



class DumbExplorationWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        params.append({'name': 'q', 'type': 'float', \
                    'limits': (1, float('inf')), 'value': 10, 'default': 10})
        
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        self.box.sample_box = DumbExploration(
                self.box.sample_box, #č rekurze)
                q=self.q)



class QTriWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        params.append({'name': 'convex_hull_degree', 'type': 'int', \
                    'title': "facet degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Degree of Grundmann-Möller cubature"})
        params.append({'name': 'tri_degree', 'type': 'int', \
                    'title': "simplex degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Degree of Grundmann-Möller cubature"})
        
#        params.append({'name': 'potential_mode', 'type': 'int', 'value': 2, 
#                    'title': "potential mode", 
#                    'limits': (0, float('inf')),
#                    'tip': "0: use rating, \n1: use pure potential, \nn: use pure ponetial for every n-th node"})
                    
        params.append({'name': 'circumcenters_only', 'title': "circumcenters only", 
                    'type': 'bool', 'value': False, 
                    'tip': "Do not treat integration nodes as candidates"})   
                    
        params.append({'name': 'store_candidates_metainformation', 'title': "store candidates metainformation", 
                    'type': 'bool', 'value': False, 
                    'tip': "True allows introspection"})   
                    
#        params.append({'name': 'weighted_entropy', 'title': "weighted entropy", 
#                    'type': 'bool', 'value': True, 
#                    'tip': "Choose to calculate agressive (weighted) or simple (vertex) entropy"})   
                    
        params.append({'name': 'q', 'type': 'float', 
                    'limits': (0, float('inf')), 'value': 1, 'default': self.box.nvar,
                    'tip': '"Agressiveness" of exploration'})
                    
        q = np.log(self.box.nvar) / self.box.nvar**2
        params.append({'name': 'psi_q', 'type': 'float', \
                    'title': "psi_q", \
                    'limits': (0, 1), 'value': 0.5, 'default': q,\
                    'tip': "Used for psi potential calculation"})   
        
        
        params.append({'name': 'holydays', 'type': 'int', \
                    'title': "holydays",  \
                    'limits': (0, float('inf')), 'value': 0, 'default': 10,\
                    'tip': "Every n-th point is an screening point. 0 to switch off"})
        
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        self.box.sample_box = QTri(
                self.box.sample_box, #č rekurze)
                self.convex_hull_degree,
                self.tri_degree,
                #potential_mode=self.potential_mode,
                store_candidates_metainformation=self.store_candidates_metainformation,
                q=self.q,
                psi_q=self.psi_q,
                circumcenters_only=self.circumcenters_only,
                holydays=self.holydays
                )



class CircumTriWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        tschemes = schemes.get_tn_keys(self.box.nvar)
        params.append({'name': 'scheme', 'type': 'list', \
                    'title': "cubature scheme", \
                     'values': tschemes, 'value': 'stroud_tn_2_1b'})
        params.append({'name': 'degree', 'type': 'int', \
                    'title': "degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Used only with Grundmann-Möller and Silvester cubaturas"})
        
#        params.append({'name': 'potential_mode', 'type': 'int', 'value': 2, 
#                    'title': "potential mode", 
#                    'limits': (0, float('inf')),
#                    'tip': "0: use rating, \n1: use pure potential, \nn: use pure ponetial for every n-th node"})
                    
        params.append({'name': 'circumcenters_only', 'title': "circumcenters only", 
                    'type': 'bool', 'value': False, 
                    'tip': "Do not treat integration nodes as candidates"})   
                    
        params.append({'name': 'store_candidates_metainformation', 'title': "store candidates metainformation", 
                    'type': 'bool', 'value': False, 
                    'tip': "True allows introspection"})   
                    
#        params.append({'name': 'weighted_entropy', 'title': "weighted entropy", 
#                    'type': 'bool', 'value': True, 
#                    'tip': "Choose to calculate agressive (weighted) or simple (vertex) entropy"})   
                    
        params.append({'name': 'q', 'type': 'float', 
                    'limits': (0, float('inf')), 'value': 1, 'default': self.box.nvar,
                    'tip': '"Agressiveness" of exploration'})
                    
        q = np.log(self.box.nvar) / self.box.nvar**2
        params.append({'name': 'psi_q', 'type': 'float', \
                    'title': "psi_q", \
                    'limits': (0, 1), 'value': 0.5, 'default': q,\
                    'tip': "Used for psi potential calculation"})   
        
        params.append({'name': 'shell_budget', 'type': 'int', \
                    'title': "shell budget",  \
                    'limits': (1, float('inf')), 'value': 1000, 'default': 1000,\
                    'tip': "Number of annulus candidates"})
        
        params.append({'name': 'holydays', 'type': 'int', \
                    'title': "holydays",  \
                    'limits': (0, float('inf')), 'value': 10, 'default': 10,\
                    'tip': "Every n-th point is sacrificed to the greatest safe simplex. 0 to switch off"})
        
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        p = self.param.child('scheme')
        p.sigValueChanged.connect(self.scheme_changed)
        
        #č podle implicitně nastavené scheme
        self.param.child('degree').hide()
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
    def scheme_changed(self, parameter):
        scheme = parameter.value()
        if scheme in ('Grundmann-Möller', 'Silvester open', 'Silvester closed'):
            self.param.child('degree').show()
        else:
            self.param.child('degree').hide()
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        scheme = schemes.get_tn_scheme(self.scheme, self.box.nvar, self.degree)
        
        
        self.box.sample_box = CircumTri(
                self.box.sample_box, #č rekurze)
                scheme,
                #potential_mode=self.potential_mode,
                store_candidates_metainformation=self.store_candidates_metainformation,
                q=self.q,
                psi_q=self.psi_q,
                shell_budget=self.shell_budget,
                circumcenters_only=self.circumcenters_only,
                holydays=self.holydays
                )



class CirQTriWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        tschemes = schemes.get_tn_keys(self.box.nvar)
        params.append({'name': 'scheme', 'type': 'list', \
                    'title': "cubature scheme", \
                     'values': tschemes, 'value': 'stroud_tn_3_6b'})
        params.append({'name': 'degree', 'type': 'int', \
                    'title': "degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Used only with Grundmann-Möller and Silvester cubaturas"})
        
        params.append({'name': 'convex_hull_degree', 'type': 'int', \
                    'title': "facet degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Degree of Grundmann-Möller cubature"})
                    
        params.append({'name': 'q', 'type': 'float', 
                    'limits': (0, float('inf')), 'value': 1, 'default': self.box.nvar,
                    'tip': '"Agressiveness" of exploration'})
        
        params.append({'name': 'holydays', 'type': 'int', \
                    'title': "holydays",  \
                    'limits': (0, float('inf')), 'value': 0, 'default': 10,\
                    'tip': "Every n-th point is an screening point. 0 to switch off"})
        
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        p = self.param.child('scheme')
        p.sigValueChanged.connect(self.scheme_changed)
        
        #č podle implicitně nastavené scheme
        self.param.child('degree').hide()
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
    def scheme_changed(self, parameter):
        scheme = parameter.value()
        if scheme in ('Grundmann-Möller', 'Silvester open', 'Silvester closed'):
            self.param.child('degree').show()
        else:
            self.param.child('degree').hide()
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        scheme = schemes.get_tn_scheme(self.scheme, self.box.nvar, self.degree)
        
        self.box.sample_box = CirQTri(
                self.box.sample_box, #č rekurze)
                scheme,
                convex_hull_degree=self.convex_hull_degree,
                q=self.q,
                holydays=self.holydays
                )




class DumbDiceBoxWidget(QtWidgets.QWidget):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent)
        
    def setup_box(self):
        # do the thing
        self.box.sample_box = DiceBox(self.box.sample_box)


class RazitkoWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        tschemes = schemes.get_tn_keys(self.box.nvar)
        params.append({'name': 'scheme', 'type': 'list', \
                    'title': "cubature scheme", \
                     'values': tschemes, 'value': tschemes[0]})
        params.append({'name': 'degree', 'type': 'int', \
                    'title': "degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Used only with Grundmann-Möller and Silvester cubaturas"})
        
        params.append({'name': 'tri_space', 'type': 'list', 'value': 'G', \
                    'title': "triangulation space", 'values': spaces})
        params.append({'name': 'tree_space', 'type': 'list', 'value': 'None', \
                    'title': "tree (potential) space", 'values': spaces+['None'], \
                    'tip': """Space where distances (and densities) are 
                    calculated. 'None' means triangulation space will be used"""})
        params.append({'name': 'sampling_space', 'type': 'list', 'value': 'None', \
                    'title': "sampling space", 'values': spaces+['None'], \
                    'tip': """Space where convex hull is being integrated. 
                    'None' means triangulation space will be used"""})
        params.append({'name': 'kechato_space', 'type': 'list', 'value': 'U', \
                    'title': "kechato space", 'values': spaces, \
                    'tip': "Only used with ksee potential"})
        params.append({'name': 'potential', 'type': 'list', 'value': 'q_psee', \
                    'title': "potential", \
                    'values': potentials})
        q = np.log(self.box.nvar) / self.box.nvar**2
        params.append({'name': 'q', 'type': 'float', \
                    'title': "q", \
                    'limits': (0, 1), 'value': q, 'default': q,\
                    'tip': "Only used with q_psee potential"})   
        params.append({'name': 'p_norm', 'type': 'float', \
                    'title': "p norm", \
                    'limits': (1, float('inf')), 'value': 2, 'default': np.inf,
                    'tip': "Space metric - used for distance calculations"})   
        params.append({'name': 'budget', 'type': 'int', \
                    'title': "Shull budget",  \
                    'limits': (1, float('inf')), 'value': 1000, 'default': 1000,\
                    'tip': "Number of nodes for outside integration"})
        params.append({'name': 'LHS_correction', \
                    'title': "LHS-like nodes placement",\
                    'type': 'bool', 'value': False })
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        scheme = schemes.get_tn_scheme(self.scheme, self.box.nvar, self.degree)
        
        if self.tree_space == 'None':
            tree_space = None
        else:
            tree_space = self.tree_space
            
        if self.sampling_space == 'None':
            sampling_space = None
        else:
            sampling_space = self.sampling_space
            
#        try:
#            stm_filename = self.box.filename + '_stm'
#        except AttributeError:
#            stm_filename = None
        
        
        #č implicitně vložime bombonjeru
        self.box.samplebox.sampled_plan = CandyBox(self.box.f_model())
        # do the thing
        self.box.sample_box = Razitko(
                self.box.sample_box, #č rekurze)
                scheme,
                tri_space=self.tri_space,
                tree_space=tree_space,
                sampling_space=sampling_space,
                kechato_space=self.kechato_space, 
                potential=self.potential, 
                q=self.q,
                p_norm=self.p_norm,
                budget=self.budget,
                LHS_correction=self.LHS_correction,
                #stm_filename=stm_filename,
                design=None
                )




class GoalWidget(pg.parametertree.ParameterTree):
    def __init__(self, wt, parent=None):
        self.box = wt
        super().__init__(parent=parent, showHeader=False)
        self._set_param()
        self.setParameters(self.param, showTop=False)
    
    def _set_param(self):
        params = list()
        tschemes = schemes.get_tn_keys(self.box.nvar)
        params.append({'name': 'scheme', 'type': 'list', \
                    'title': "cubature scheme", \
                     'values': tschemes, 'value': tschemes[0]})
        params.append({'name': 'degree', 'type': 'int', \
                    'title': "degree",\
                    'limits': (0, float('inf')), 'value': 5, 'default': 5,\
                    'tip': "Used only with Grundmann-Möller and Silvester cubaturas"})
        
        params.append({'name': 'tri_space', 'type': 'list', 'value': 'G', \
                    'title': "triangulation space", 'values': spaces})
        params.append({'name': 'tree_space', 'type': 'list', 'value': 'None', \
                    'title': "tree (potential) space", 'values': spaces+['None'], \
                    'tip': """Space where distances (and densities) are 
                    calculated. 'None' means triangulation space will be used"""})
        params.append({'name': 'kechato_space', 'type': 'list', 'value': 'U', \
                    'title': "kechato space", 'values': spaces, \
                    'tip': "Only used with ksee potential"})
        params.append({'name': 'potential', 'type': 'list', 'value': 'q_psee', \
                    'title': "potential", \
                    'values': potentials})
        q = np.log(self.box.nvar) / self.box.nvar**2
        params.append({'name': 'q', 'type': 'float', \
                    'title': "q", \
                    'limits': (0, 1), 'value': q, 'default': q,\
                    'tip': "Only used with q_psee potential"})   
        params.append({'name': 'p_norm', 'type': 'float', \
                    'title': "p norm", \
                    'limits': (1, float('inf')), 'value': 2, 'default': np.inf,
                    'tip': "Space metric - used for distance calculations"})   
        params.append({'name': 'shell_budget', 'type': 'int', \
                    'title': "shell budget",  \
                    'limits': (1, float('inf')), 'value': 1000, 'default': 1000,\
                    'tip': "Number of annulus candidates"})
        
        params.append({'name': 'outer_budget', 'type': 'int', \
                    'title': "Outer budget",  \
                    'limits': (1, float('inf')), 'value': 100, 'default': 100,\
                    'tip': "Number of candidates nodes outside of circumscribed d-ball"})
        
        params.append({'name': 'LHS_correction', \
                    'title': "LHS-like nodes placement",\
                    'type': 'bool', 'value': False })
        
        ### Create tree of Parameter objects
        # I don't know why that signals do not work for me
        # Only sigTreeStateChanged works, but I don't want to struggle with it
        # May be I'll report the issue 
        #self.param.sigValueChanged.connect(self.param_changed)
        #self.param.sigValueChanging.connect(self.param_changing)
        self.param = pg.parametertree.Parameter.create(name='params', type='group', children=params)
        
        #č branima sa rekurzii
        #оӵ рекурзилы пезьдэт!
        self.param_values = self.param.getValues()
        
    def __getattr__(self, attr):
        #č na teoreticky možnou rěkurziju vykašleme
        #оӵ рекурзия уз луы
        return self.param_values[attr][0]
        
    def setup_box(self):
        #č to je důležité! __getatr__ odsaď bere hodnoty
        self.param_values = self.param.getValues()
        
        scheme = schemes.get_tn_scheme(self.scheme, self.box.nvar, self.degree)
        
        if self.tree_space == 'None':
            tree_space = None
        else:
            tree_space = self.tree_space
            
        try:
            stm_filename = self.box.filename + '_stm'
        except AttributeError:
            stm_filename = None
        
        #č implicitně vložime bombonjeru
        self.box.samplebox.sampled_plan = CandyBox(self.box.f_model())
        self.box.sample_box = Goal(
                self.box.sample_box, #č rekurze)
                scheme,
                tri_space=self.tri_space,
                tree_space=tree_space,
                kechato_space=self.kechato_space, 
                potential=self.potential, 
                q=self.q,
                p_norm=self.p_norm,
                shell_budget=self.shell_budget,
                outer_budget=self.outer_budget,
                LHS_correction=self.LHS_correction,
                stm_filename=stm_filename,
                )
        

class SetupDiceBoxWidget(pg.LayoutWidget):
    def __init__(self, wt, parent=None):
        self.box = wt
        #č nejdřív vytvořiť apku
        self.app = pg.mkQApp("WellMet")
        # 
        super().__init__(parent)
        
        self.setup()
        
        #č zobraziť
        self.show()
        #č a spustit smyčku
        self.app.exec_()
        
    def setup(self):
        self.setWindowTitle("Set up sequential algorithm")
        
        self.setup_tabs()
        self.addWidget(self.tab_widget, row=0, col=0, rowspan=1, colspan=2)
        
        self.btn_skip = QtWidgets.QPushButton('Skip')
        self.addWidget(self.btn_skip, row=1, col=0)
        self.btn_skip.clicked.connect(self.skip)
        
        self.btn_choose = QtWidgets.QPushButton('Create')
        self.addWidget(self.btn_choose, row=1, col=1)
        self.btn_choose.clicked.connect(self.create)
        
    def skip(self):
        self.app.quit()
        
    def create(self):
        tab_index = self.tab_widget.currentIndex()
        self.tabs[tab_index].setup_box()
        self.app.quit()
        
    def setup_tabs(self):
        # Initialize tab screen
        self.tab_widget = QtWidgets.QTabWidget(self)
        #č šlo by to, samozřejmně, i přes Qt
        #č ale stejně musíme reference explicitně ukladat 
        #č aby je nám Python nehodil
        self.tabs = []


        box_widget = CirQTriWidget(self.box, self)
        self.tab_widget.addTab(box_widget, "CirQTri")
        self.tabs.append(box_widget)        
        




#!/usr/bin/env python
# coding: utf-8

"""
č chce se ukladat meta infa k bodíkům.
č CandyBox nepodporuje zadání jednoduchých atributů
č pro celou sadů bodíků. Navíc je těžký a hrozně ošklivý.
č CandyNodes je myšleno jako jednoduchá alternativa
č pro použití uvnitř skříňky, která nejde do vzorků.
"""

class CandyNodes:
    #č to je vše co má samotné CandyNodes
    #č žádný další majetek nepozoruji
    __slots__ = 'sampling_plan', 'attrs', 'kwargs'
    
    def __new__(cls, sample_object, attrs=None, /, **kwargs):
        """
        č Jedname tvrdě - není-li vstup konzistentní, 
        č tak tenhle box vůbec nevytvaříme
        """
        cb = super().__new__(cls)
        cb.sampling_plan = sample_object
        cb.kwargs = kwargs
        if attrs is None:
            cb.attrs = {}
        else:
            cb.attrs = attrs
        
        
        
#        for key, value in kwargs.items():
#            if hasattr(value, '__getitem__') and hasattr(value, "__len__"):
#                cb.kwargs[key] = value
#            else:
#                cb.attrs[key] = value
        
        
        if cb.consistency_check():
            return cb
        else:
            raise ValueError("Sample and given values hasn't the same length. Zkrátka, do sebe nepatří")
            
            
        
    def __repr__(cb):
        return 'CandyNodes(%s, %s, **%s)' % (repr(cb.sampling_plan),\
                                         repr(cb.attrs), repr(cb.kwargs))
        
    def __len__(cb):
        return len(cb.sampling_plan)
        
        
    def __call__(cb, *args, **kwargs):
        # Call meaning is different for underlaying f_models and upper Boxes,
        # however SampleBox expect f_model under.
        # Also CandyNodes tryes to be "better f_model", 
        # so should behave similarly
        f = cb.sampling_plan(*args, **kwargs)
        return CandyNodes(f)
            
    
    def __getitem__(cb, subscript):
        #č subscript vratí nový instance
        #č s nezávislými slovníky.
        #č sice numpy matice uvnitř zůstanou závislé
        #č ale atributy CandyNodes - nikoliv.
        cn = CandyNodes(cb.sampling_plan[subscript])
        #ё нам можна
        cn.attrs.update(cb.attrs)
        #č Sice nechavám na uživateli co vloží do slovníku 
        #č a jak se to bude slajsiť, ale pořad očekavám ideomatické chování
        #č Ošetřuji případ jednoduchého indexu (array[i]), 
        #č který z kontejneru vratí osamelý prvek
        #č a slice, který zas vratí odříznutý kontejněr.
        #č Numpy ještě může brát bulové pole jako index
        if isinstance(subscript, slice) or hasattr(subscript, "__len__"):
            dest_dict = cn.kwargs
        else:
            dest_dict = cn.attrs
        for key, data in cb.kwargs.items():
            dest_dict[key] = data[subscript]
        return cn
        
        
    def __setattr__(cb, key, value):
        try:
            super().__setattr__(key, value)
        except AttributeError:
            #č zde musím hlídat aby attrs a kwargs neobsahovaly stejné klíče
            if hasattr(value, '__getitem__') and hasattr(value, "__len__"):
                if len(cb) != len(value):
                    raise ValueError("Sample and given values hasn't the same length. Zkrátka, do sebe nepatří")
                cb.kwargs[key] = value
                cb.attrs.pop(key, None)
            else:
                cb.attrs[key] = value
                cb.kwargs.pop(key, None)
        
    def __delattr__(cb, key):
        cb.kwargs.pop(key, None)
        cb.attrs.pop(key, None)
        
        
    def __getattr__(cb, attr):
        #č candybox, třeba, nikdy jsem nepoužil
        if attr == 'candynodes':
            return cb
            
        #č hledáme obraceně
        #č nejdřív se zeptame f_model,
        #č teprve když ten nič nemá  
        #č mrkneme atribut u sebe
        try:
            return getattr(cb.sampling_plan, attr)
        except AttributeError:
            if attr in cb.kwargs:
                value = cb.kwargs[attr]
                if len(cb)==len(value):
                    return value
                else:
                    #ё у нас есть дата, но мы их вам недадим)
                    #č zde nechť bude KeyError
                    raise KeyError("CandyNodes: well, we have some data, but they are not consistent, so we haven't")
            elif attr in cb.attrs:
                return cb.attrs[attr]
            else: #č implicitně slovníky hazej KeyError, kterej nechcem
                raise AttributeError
            
            
    def set_attrs(cb, *args, **kwargs):
        #č Je tu nepřijemnost pro uživatele třídy.
        #č Pokud kwargy jíž obsahují klíč,
        #č tak pořád se bude vrácet on protože má přednost.
        #č Nechcu to řešit.
        cb.attrs.update(*args, **kwargs) 
        
        
     
     #č Funkce musí být definována, 
     #č jinak se automaticky zavolá add_sample f-modelu,
     #č a o tom ani nedozvíme
    def add_sample(cb, input, **kwargs):
        #č výkon může zaviset na tom, co posíláme f_modelu: 
        #č f_model, nebo bůhvíco mu neznámého
        #č navíc, uvnitř f_model vůbec nemusí být
        if hasattr(input, 'candynodes'): #sweety_input:
            input = input.sampling_plan
        cb.sampling_plan.add_sample(input, **kwargs)
        #č co mám dělat s kwargy?
        #č doplňovat? Nemám jak - nehlídám a netrvám na numpy
        #č může být v datech cokoli
        #č Hazet chybu? Vyhozovat kwargy?
        if cb.kwargs:
            cb.kwargs.clear()
        
        
    
    #č netuším, kdo to potřebuje v novém kódě
    def new_sample(cb, input=None, **kwargs):
        #č pokud input není, vrácíme samy sobě.
        if input is None:
            return CandyNodes(cb.sampling_plan.new_sample(**kwargs))
            
        #č jinak konvertujem vstup na naše rozdělení
        #č čo to je za vstup?
        if hasattr(input, 'candynodes'): #sweety_input:
            cn = CandyNodes(cb.sampling_plan.new_sample(input.sampling_plan, **kwargs))
            cn.attrs.update(input.attrs)
            cn.kwargs.update(input.kwargs)
            return cn
            
        #č nesladký vstup
        return CandyNodes(cb.sampling_plan.new_sample(input, **kwargs))
        
        
        
    def consistency_check(cb):
        # řvat na celé město nebudeme
        sample_len = len(cb.sampling_plan)
        return all(sample_len == len(cb.kwargs[key]) for key in cb.kwargs)



#!/usr/bin/env python
# coding: utf-8


"""
 cs: 
 

 en: 
 data should be pandas compatible, i.e. 
 nsim, nvar = data.shape

g_model returns SampleBox object, which actually contains:
1. what-was-on-input
2. values of perfomance function
AND 3. so called gm_signature - some string to identify data
and do not let them be mixed up.
Currently WhiteBox treat as gm_signature __name__ attribute (free functions has it),
otherwise repr(). Classes supposed to define __repr__ function for correct work.

Fence off!
Some of performance functions (g_models) are able to draw
failure region boundary, in this case 
.get_2D_R_boundary(nrod, xlim, ylim) is defined.
    nrod - number of "rods" in "fencing"
    xlim, ylim describes your viewport, plotting terminal, whatever
    g_model uses these parameters only for inspiration,
    g_model is allowed to ignore them
    xlim = (xmin, xmax)
    ylim = (ymin, ymax)
    
    returns tuple (or list) of R samples
    
    
    
"""

import numpy as np
from .f_models import Ingot
from .samplebox import SampleBox


class GetQuadrantBoundary2D:
    """
    sebemenší pomocná třida pro vykreslení hranici L tvaru
    """
    def __init__(self, center_point=(0,0), quadrant='I'):
        """
        quadrants also сэрегъёс-compatible
        #### CORNERS 2D #####
        # print(сэрегъёс)
        # numbering:
        #    2  |  3
        #  -----|-----
        #    0  |  1
        """
        self.center_point = center_point
        self.quadrant = quadrant
        
    def __call__(self, nrod=100, xlim=(-5,5), ylim=(-5,5)):
        xc, yc = self.center_point
        xmin = min(*xlim, xc-1)
        xmax = max(*xlim, xc+1)
        ymin = min(*ylim, yc-1)
        ymax = max(*ylim, yc+1)
            
        nrod = int(nrod/2)
            # mně nic hezčího prostě nenapadá(
        if self.quadrant in ('I', 3):
            xbound = np.append(np.full(nrod, xc), np.linspace(xc, xmax, nrod, endpoint=True))
            ybound = np.append(np.linspace(ymax, yc, nrod, endpoint=True), np.full(nrod, yc))
        elif self.quadrant in ('II', 2):
            xbound = np.append(np.linspace(xmin, xc, nrod, endpoint=True), np.full(nrod, xc))
            ybound = np.append(np.full(nrod, yc), np.linspace(yc, ymax, nrod, endpoint=True))
        elif self.quadrant in ('III', 0):
            xbound = np.append(np.linspace(xmin, xc, nrod, endpoint=True), np.full(nrod, xc))
            ybound = np.append(np.full(nrod, yc), np.linspace(yc, ymin, nrod, endpoint=True))
        else: # self.quadrant in ('IV', 1):
            xbound = np.append(np.full(nrod, xc), np.linspace(xc, xmax, nrod, endpoint=True))
            ybound = np.append(np.linspace(ymin, yc, nrod, endpoint=True), np.full(nrod, yc))
        
    
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.vstack((xbound, ybound)).T
        # tuple of tuple
        return (Ingot(bound_R),)
        
        

def get_R_coordinates(input_sample, envar=0):
    """
    Tohle je pomocná funkce, vrácí g_modelům numpy 2d pole s daty
    
    envar - zadavejte, pokud chcete zkontrolovat 
    počet náhodných proměnných
    envar like estimated number of variables
    """
    # is it sample object?
    try:
        if envar > 0 and input_sample.nvar != envar:
            raise ValueError('%sD data expected, but %sD sample given'% (envar, input_sample.nvar))
        else:
            return input_sample.R
        
    # it is not an sample object, 
    # but maybe numpy can handle this?
    except AttributeError:
        sample = np.atleast_2d(np.array(input_sample))
        # invar like input number of variables
        nsim, invar = sample.shape
        if envar > 0 and invar != envar:
            raise ValueError('%sD data expected, but %sD sample given'% (envar, invar))
        else:
            return sample




class Linear_nD:
    """
    Class takes for inicialization tuple of betas 
    Betas are coeffitients in sense of Regression Analysis
    
     g= a*X1 + b*X2 + c
     becames
     gm = Linear_nD(betas=(a,b,c))
     gm(samples)
     gm.get_2D_R_boundary(nrod, xlim) returns
     xbounds a ybounds zabalené do tuplu, ty zabalené do listu
    """
    
    def __init__(self, betas):
        self._betas = betas
        
        # sign
    def __repr__(self):
        return 'Linear_nD(%s)' % repr(self._betas)
        
    def __call__(self, input_sample):
        selfnvar = len(self._betas)-1
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, selfnvar)
        # teďkom zasahujeme přímo do tohoto pole
        sim = sample.copy()
        for i in range(selfnvar):
            sim[:,i] = sim[:,i]*self._betas[i]
        g = np.sum(sim, axis=1) + self._betas[-1]
        return SampleBox(input_sample, g, repr(self))
    
    
    
    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        xbound = np.linspace(xlim[0], xlim[1], nrod, endpoint=True)
    
        selfnvar = len(self._betas)-1
        # g= a*X1 + b*X2 + 0*X3 + 0*X4 + ... + c
        a = self._betas[0]
        b = self._betas[1]
        c = self._betas[-1]
        
        # 1D je spíše vtip
        if selfnvar == 1:
            return (-c/a)
        else:
            # sample compatible
            # малы транспонировать кароно? Озьы кулэ!
            bound_R = np.array((xbound, -c/b + (-a/b)*xbound)).T
            # tuple of tuple
            return (Ingot(bound_R),)
        
        
class X1:
    """
    suma velicin plus beta*sqrt(Nvar. )
    Pro IID Gaussian ma tohle ind. spol. beta = beta
    The same as Linear_nD, but defined via 
    beta in sense of reliability index
    """
    def __init__(self, beta_exact):
        self._beta_exact = beta_exact
        
    # sign
    def __repr__(self):
        return 'X1(%s)' % repr(self._beta_exact)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = -sample[:, 0] + self._beta_exact
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), ylim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        bound_R = np.empty((nrod, 2))
        bound_R[:, 0] = self._beta_exact
        bound_R[:, 1] = np.linspace(ylim[0], ylim[1], nrod, endpoint=True)
        # tuple of tuple
        return (Ingot(bound_R),)
        
        
class AbsX1:
    """
    suma velicin plus beta*sqrt(Nvar. )
    Pro IID Gaussian ma tohle ind. spol. beta = beta
    The same as Linear_nD, but defined via 
    beta in sense of reliability index
    """
    def __init__(self, beta_exact):
        self._beta_exact = beta_exact
        
    # sign
    def __repr__(self):
        return 'X1(%s)' % repr(self._beta_exact)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = -np.abs(sample[:, 0]) + self._beta_exact
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), ylim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        bound_R = np.empty((nrod, 2))
        bound_R[:, 0] = self._beta_exact
        bound_R[:, 1] = np.linspace(ylim[0], ylim[1], nrod, endpoint=True)
        # tuple of tuple
        return (Ingot(bound_R), Ingot(-bound_R))
        
        
class Z_sum:
    """
    suma velicin plus beta*sqrt(Nvar. )
    Pro IID Gaussian ma tohle ind. spol. beta = beta
    The same as Linear_nD, but defined via 
    beta in sense of reliability index
    """
    def __init__(self, nvar, beta_exact):
        self._nvar = nvar
        self._beta_exact = beta_exact
        
    # sign
    def __repr__(self):
        return 'Z_sum(%s, %s)' % (repr(self._nvar), repr(self._beta_exact))
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, self._nvar)
        g = np.sum(sample, axis=1) + self._beta_exact * np.sqrt(self._nvar) 
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        xbound = np.linspace(xlim[0], xlim[1], nrod, endpoint=True)
    
        # g= a*X1 + b*X2 + 0*X3 + 0*X4 + ... + c
        a = 1
        b = 1
        c = self._beta_exact * np.sqrt(self._nvar) 
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((xbound, -c/b + (-a/b)*xbound)).T
        # tuple of tuple
        return (Ingot(bound_R),)

      
        
class Z_prod:
    """
    soucin velicin plus nějaká konstanta 
    # g= s * (X1 * X2 * X3 * X4 + c )
    """
    # tenhle model ani nvar si neukladá, tohle vůbec neřeší
    def __init__(self, const=0, sign=1):
        self._const = const
        self._sign = sign
        
    # sign
    def __repr__(self):
        return 'Z_prod(%s, %s)' % (self._const, self._sign)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = self._sign * (np.prod(sample, axis=1) + self._const)
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        # g= X1 * X2 + c
        #
        # a^2 = 2*X^2 
        # a=b= X * sqrt(2)
        # a^2 = 2*c
        # r = a*b / np.sqrt(b**2 * np.cos(phi)**2 - a**2 * np.sin(phi)**2)
        
        c = self._const
        _c = np.sign(c)
        #č náš oblibený trik - hranici nakreslime pomoci polárních souřádnic
        phi = np.linspace(0.25*np.pi, (0.25+_c/2)*np.pi, nrod , endpoint=False)[1:]
        r = np.sqrt(2*c / (np.sin(phi)**2 - np.cos(phi)**2))
        bound_x_left = r * np.cos(phi+np.pi/4)
        bound_y_left = r * np.sin(phi+np.pi/4)
        
        phi = np.linspace(-0.75*np.pi, (_c/2-0.75)*np.pi, nrod , endpoint=False)[1:]
        r = np.sqrt(2*c / (np.sin(phi)**2 - np.cos(phi)**2))
        bound_x_right = r * np.cos(phi+np.pi/4)
        bound_y_right = r * np.sin(phi+np.pi/4)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R_left = np.array((bound_x_left, bound_y_left)).T
        bound_R_right = np.array((bound_x_right, bound_y_right)).T
        # tuple of samples
        return (Ingot(bound_R_left), Ingot(bound_R_right))
             
        

class Z_hyperbola:
    """
    soucin velicin plus nějaká konstanta 
    # g= s * (X1 * X2 * X3 * X4 + c )
    """
    # tenhle model ani nvar si neukladá, tohle vůbec neřeší
    def __init__(self, const=0):
        self._const = const
        
    # sign
    def __repr__(self):
        return 'Z_hyperbola(%s)' % self._const
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = self._const - np.all(sample > 0, axis=1) * np.prod(sample, axis=1)
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        # g= X1 * X2 + c
        #
        # a^2 = 2*X^2 
        # a=b= X * sqrt(2)
        # a^2 = 2*c
        # r = a*b / np.sqrt(b**2 * np.cos(phi)**2 - a**2 * np.sin(phi)**2)
        
        c = -self._const
        _c = -1
        #č náš oblibený trik - hranici nakreslime pomoci polárních souřádnic
        phi = np.linspace(0.25*np.pi, (0.25+_c/2)*np.pi, nrod , endpoint=False)[1:]
        r = np.sqrt(2*c / (np.sin(phi)**2 - np.cos(phi)**2))
        bound_x_left = r * np.cos(phi+np.pi/4)
        bound_y_left = r * np.sin(phi+np.pi/4)
        
#        phi = np.linspace(-0.75*np.pi, (_c/2-0.75)*np.pi, nrod , endpoint=False)[1:]
#        r = np.sqrt(2*c / (np.sin(phi)**2 - np.cos(phi)**2))
#        bound_x_right = r * np.cos(phi+np.pi/4)
#        bound_y_right = r * np.sin(phi+np.pi/4)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R_left = np.array((bound_x_left, bound_y_left)).T
        #bound_R_right = np.array((bound_x_right, bound_y_right)).T
        # tuple of samples
        return (Ingot(bound_R_left), ) #Ingot(bound_R_right))


class Z_min:
    """
    min velicin plus nějaká konstanta 
    # g= min(X1, X2, X3, X4) + c
    """
    def __init__(self, const):
        self._const = const
        self.get_2D_R_boundary = GetQuadrantBoundary2D(center_point=(-const,-const), quadrant='I')
        
    # sign
    def __repr__(self):
        return 'Z_min(%s)' % repr(self._const)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = np.min(sample, axis=1) + self._const
        return SampleBox(input_sample, g, repr(self))
    
    # samotná "volaná" se určí v __init__
    # trik aby se nezabindila
    @staticmethod
    def get_2D_R_boundary(): return None
        

class Z_max:
    """
    max velicin plus nějaká konstanta 
    # g= max(X1, X2, X3, X4) + c
    """
    def __init__(self, const):
        self._const = const
        self.get_2D_R_boundary = GetQuadrantBoundary2D(center_point=(-const,-const), quadrant='III')
        
    # sign
    def __repr__(self):
        return 'Z_max(%s)' % repr(self._const)
        
    def __call__(self, input_sample):
        #č očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = np.max(sample, axis=1) + self._const
        return SampleBox(input_sample, g, repr(self))
    
    #č samotná "volaná" se určí v __init__
    #č trik aby se nezabindila
    @staticmethod
    def get_2D_R_boundary(): return None


class Z_sumexp:
    """
    """
    def __init__(self, const):
        self._const = const
        
    # sign
    def __repr__(self):
        return 'Z_sumexp(%s)' % repr(self._const)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = np.sum(np.exp(-(sample**2)), axis=1) + self._const
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), ylim=(-5,5)):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        
        def e_bound(xbound): return np.sqrt(-np.log(-self._const - np.exp(-xbound**2)))
        
        # let's mirror about (s,s) point
        # 0 = 2*e^(-s^2)+c
        # log(-c/2) = -s^2
        # s = sqrt(-log(-c/2))
        s = np.sqrt(-np.log(-self._const/2))
        xmin = min(*xlim, -s-1)
        xmax = max(*xlim, s+1)
        ymin = min(*ylim, -s-1)
        ymax = max(*ylim, s+1)
        
        xb_1eft = np.linspace(xmin, -s, nrod, endpoint=False)
        xb_right = np.linspace(xmax, s, nrod, endpoint=False)
        yb_up = np.linspace(s, ymax, nrod)
        yb_down = np.linspace(-s, ymin, nrod)
        
        
        # numerace je náhodná
        bound_R_1 = np.array((np.append(xb_1eft, -e_bound(yb_up)), np.append(e_bound(xb_1eft), yb_up))).T
        bound_R_2 = np.array((np.append(xb_1eft, -e_bound(yb_down)), np.append(-e_bound(xb_1eft), yb_down))).T
        bound_R_3 = np.array((np.append(xb_right, e_bound(yb_up)), np.append(e_bound(xb_right), yb_up))).T
        bound_R_4 = np.array((np.append(xb_right, e_bound(yb_down)), np.append(-e_bound(xb_right), yb_down))).T
        
        # sample compatible
        # tuple of samples
        return (Ingot(bound_R_1), Ingot(bound_R_2), Ingot(bound_R_3), Ingot(bound_R_4))
              
        


class S_ballSin2D:
    """
    c = 0.5 # wave amplitude in Gaussian space
    d = 3.0 # average of sine fiunction in Gaussian space
    k = 6   # number of sine waves (design points)
    """
    def __init__(self, c, d, k):
        self._c = c
        self._d = d
        self._k = k
        
    # sign
    def __repr__(self):
        return 'S_ballSin2D(%s,%s,%s)' % (self._c, self._d, self._k)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        R2 = np.sum(np.square(sample), axis=1)
        R = np.sqrt(R2)
        phi = np.arctan2(sample[:,1] , sample[:,0])  #arctan2(y,x)
        rmax = self._c * np.sin(self._k * phi) + self._d
        g = rmax - R 
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        phi = np.linspace(0, 6.283185307, nrod , endpoint=True)
        r = self._c * np.sin(self._k * phi) + self._d
        bound_x = r * np.cos(phi)
        bound_y = r * np.sin(phi)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((bound_x, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)


class Z_sumsq:
    """
    """
    def __init__(self, const):
        self._const = const
        
    # sign
    def __repr__(self):
        return 'Z_sumsq(%s)' % repr(self._const)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        g = np.sum(sample**2, axis=1) - self._const
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        phi = np.linspace(0, 2*np.pi, nrod , endpoint=True)
        r = np.sqrt(self._const)
        bound_x = r * np.cos(phi)
        bound_y = r * np.sin(phi)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((bound_x, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)



class S_ball:
    """
    Find 10 differences with Z_sumsq
    """
    def __init__(self, r):
        self._r = r
        
    # sign
    def __repr__(self):
        return 'S_ball(%s)' % repr(self._r)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        R2 = np.sum(np.square(sample), axis=1)
        g = self._r**2 - R2
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        phi = np.linspace(0, 6.283185307, nrod, endpoint=True)
        r = self._r
        bound_x = r * np.cos(phi)
        bound_y = r * np.sin(phi)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((bound_x, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)


class ConicSection:
    """
    """
    def __init__(self, l, e, teta=0, c=(0,0), sign=1):
        self._l = l
        self._e = e
        self._teta = teta
        self._c = c
        self._sign = sign
        
    # sign
    def __repr__(self):
        return 'ConicSection(%s, %s, %s, %s, %s)' % (repr(self._l), repr(self._e), repr(self._teta), repr(self._c), repr(self._sign))
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        x_i, y_i, *__ = (*sample.T,)
        x = x_i - self._c[0]
        y = y_i - self._c[1]
        phi = np.arctan2(y, x)# zde musí bejt to obracené pořádí
        
        r = np.sqrt(np.square(x) + np.square(y))
        r_bound = abs(self._l / (1 - self._e * np.cos(phi-self._teta)))
        g = self._sign * (r - r_bound)
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        if self._e == 1:
            phi = np.linspace(self._teta, 6.283185307+self._teta, nrod, endpoint=False)[1:]
        else:
            phi = np.linspace(0, 6.283185307, nrod, endpoint=True)
            
        r_bound = abs(self._l / (1 - self._e * np.cos(phi-self._teta)))
        
        bound_x = (r_bound * np.cos(phi)) + self._c[0]
        bound_y = (r_bound * np.sin(phi)) + self._c[1]
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((bound_x, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)


class Exp_P:
    """
        g = y - 1./np.exp(x)**5
    """
    def __init__(self, k=1., pow=5):
        self._k = k
        self._pow = pow
        
    # sign
    def __repr__(self):
        return 'Exp_P(%s, %s)' % (self._k, self._pow)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        x = sample[:,0]
        y = sample[:,1]
        g = y - self._k/np.exp(x)**self._pow
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        xbound = np.linspace(xlim[0], xlim[1], nrod, endpoint=True)
        
        bound_y = self._k/np.exp(xbound)**self._pow
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((xbound, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)
            
            
class Sin2D:
    """
    """
    def __init__(self, kx=-1/4., ky=-1, kxsin=5, const=5):
        self._kx = kx
        self._ky = ky
        self._kxsin = kxsin
        self._const = const
        
    # sign
    def __repr__(self):
        return 'Sin2D(%s, %s, %s, %s)' % (self._kx, self._ky, self._kxsin, self._const)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        x = sample[:,0]
        y = sample[:,1]
        g = self._kx * x + self._ky * y + np.sin(self._kxsin*x) + self._const
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        xbound = np.linspace(xlim[0], xlim[1], nrod, endpoint=True)
        
        bound_y = -(self._kx * xbound  + np.sin(self._kxsin * xbound) + self._const) / self._ky
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((xbound, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),) 



class Prod_FourBetas:
    """
    g = beta^2/2 - |x1 * x2|
    """
    def __init__(self, beta=2.0):
        self._beta = beta
        
    # sign
    def __repr__(self):
        return 'Prod_FourBetas(beta=%s)' % repr(self._beta)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)
        
        g = self._beta**2/2.0 - np.prod(np.abs(sample), axis=1)
        return SampleBox(input_sample, g, repr(self))

    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), ylim=(-5,5)):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        # zde vynechal abs(), ale úpravil znaménka dolů
        # don't ask me why. Ачим но уг тодӥськы.
        def e_bound(xbound): return self._beta**2/2 / xbound
        
        # let's mirror about (s,s) point
        # 0 = beta^2/2 - s^2
        # beta^2/2 = s^2
        # s = beta/sqrt(2)
        s = self._beta / np.sqrt(2)
        xmin = min(*xlim, -s-1)
        xmax = max(*xlim, s+1)
        ymin = min(*ylim, -s-1)
        ymax = max(*ylim, s+1)
        
        xb_1eft = np.linspace(xmin, -s, nrod, endpoint=False)
        xb_right = np.linspace(xmax, s, nrod, endpoint=False)
        yb_up = np.linspace(s, ymax, nrod)
        yb_down = np.linspace(-s, ymin, nrod)
        
        
        # numerace je náhodná. Je tu hračka se znaménky.
        bound_R_1 = np.array((np.append(xb_1eft, -e_bound(yb_up)), np.append(-e_bound(xb_1eft), yb_up))).T
        bound_R_2 = np.array((np.append(xb_1eft, e_bound(yb_down)), np.append(e_bound(xb_1eft), yb_down))).T
        bound_R_3 = np.array((np.append(xb_right, e_bound(yb_up)), np.append(e_bound(xb_right), yb_up))).T
        bound_R_4 = np.array((np.append(xb_right, -e_bound(yb_down)), np.append(-e_bound(xb_right), yb_down))).T
        
        # sample compatible
        # tuple of samples
        return (Ingot(bound_R_1), Ingot(bound_R_2), Ingot(bound_R_3), Ingot(bound_R_4))
        
        
        
class BlackSwan2D:
    """
        a = 2.0 # boundary for x1
        b = 5.0 # boundary for x2
        y = np.where(sim[:,0] <= a, sim[:,0], sim[:,1])
        # pro x1 <= a   y = x1
        # pro x1 >  a   y = x2
        g = b - y # failure for b<y
    """
    def __init__(self, a=2.0, b=5.0):
        self._a = a
        self._b = b
        if a<b:
            self.get_2D_R_boundary = GetQuadrantBoundary2D(center_point=(a, b), quadrant='I')
        
    # sign
    def __repr__(self):
        return 'BlackSwan2D(%s, %s)' % (self._a, self._b)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        y = np.where(sample[:,0] <= self._a, sample[:,0], sample[:,1])
        g = self._b - y # failure for b<y
        return SampleBox(input_sample, g, repr(self))


    # samotná "volaná" se určí v __init__
    # trik aby se nezabindila
    @staticmethod
    def get_2D_R_boundary(*args, **kwargs): 
        # jako kdyby .get_2D_R_boundary nebyla vůbec definována.
        raise AttributeError
        


class Metaballs2D:
    """
    """
    def __init__(self, const=5):
        # sebemenší parametrizace
        self._const = const
        
    # sign
    def __repr__(self):
        return 'Metaballs2D(%s)' % repr(self._const)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        x1 = sample[:,0]  
        x2 = sample[:,1]
        
        # auxiliary variables
        y1 = 4/9*(x1 + 2  )**2 + 1/25 * (x2    )**2 
        y2 = 1/4*(x1 - 2.5)**2 + 1/25 * (x2-0.5)**2 
        g = 30.0/( y1**2 + 1.0 ) + 20.0/( y2**2 + 1.0 ) - self._const
        return SampleBox(input_sample, g, repr(self))
    
    
    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        #č myšlenkou je vygenerovat rovnoměrně spoustu teček v každém směru,
        #č pro každé phi zvolit nejblížší k nule r-ko
        phi = np.linspace(0, 6.283185307, nrod , endpoint=True)
        #č obecná hranice poruchy ӧвӧл :(
        #č tyhle meze vícemené platí jen pro const=5
        r = np.linspace(4, 8, nrod//2 , endpoint=True)
        X = np.atleast_2d(r).T @ np.cos(np.atleast_2d(phi))
        Y = np.atleast_2d(r).T @ np.sin(np.atleast_2d(phi))
        # auxiliary variables
        y1 = 4/9*(X + 2  )**2 + 1/25 * (Y    )**2 
        y2 = 1/4*(X - 2.5)**2 + 1/25 * (Y-0.5)**2 
        g = 30.0/( y1**2 + 1.0 ) + 20.0/( y2**2 + 1.0 ) - self._const
        
        #č nechapu úplně přoč tam není axis=1, ale O, ale to funguje
        arg_r = np.argmin(np.abs(g), axis=0)
        bound_r = r[arg_r]
        
        bound_x = bound_r * np.cos(phi)
        bound_y = bound_r * np.sin(phi)
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((bound_x, bound_y)).T
        # tuple of samples
        return (Ingot(bound_R),)
        
        
class Logistic2D:
    """
    """
    def __init__(self, c1=5, c2=4, easy_version=True):
        # sebemenší parametrizace
        self._c1 = c1
        self._c2 = c2
        self.easy_version = easy_version
        
        # For both versions
        
        self.get_2D_R_boundary = GetQuadrantBoundary2D(center_point=(c1,-c2), quadrant='II')
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        
        x1 = sample[:,0]  
        x2 = sample[:,1]
        
        # auxiliary variables
        y1 = self._c1 - x1
        y2 = self._c2 + x2
        y3 = 1.0/(1+np.exp(-2.0*y2)) - 0.5
        
        if self.easy_version:
            g = np.minimum(y1,y2)  # easy version for SuS
        else:
            g = np.minimum(y1,y3)  # difficult version for SuS
        return SampleBox(input_sample, g, repr(self))
    
    def __repr__(self):
        return 'Logistic2D(%s, %s, easy_version=%s)'%(self._c1, self._c2, self.easy_version)
    
    def pf_expression(self, f_model):
        """
        We trying to say how to calculate pf
        to someone, who will know actual distribution 
        """
        a = f_model.marginals[0].sf(self._c1)
        b = f_model.marginals[1].cdf(-self._c2)
        # subtract the twice calculated intersection
        return a + b - a*b, 'exact solution'
    
    @staticmethod
    def get_2D_R_boundary(): return None
    
        
        
class CosExp2D:
    """
    """
    def __init__(self, s=5):
        # sebemenší parametrizace
        self._s = s
        
    # sign
    def __repr__(self):
        return 'CosExp2D(s=%s)' % repr(self._s)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)

        # auxiliary variables
        s = self._s
        # g = cos((np.exp(-xm-s  ))*xm)   * np.exp(-(x +s  )/3)
        g = np.cos( ( np.exp(-sample[:,0] - s ) )*sample[:,0])   * np.exp( -(sample[:,0] + s  )/3 )        
        return SampleBox(input_sample, g, repr(self))
        
    def pf_expression(self, f):
        xs, rad_max = self._get_x()
        sign = np.sign(np.cos(rad_max))
        #sign = (k_max + 1) // 2 - 1
        xs.reverse()
        print(xs)
        cdfs = f.marginals[0].cdf(xs)
        print(cdfs)
        if sign > 0:
            pf = 0
        else:
            pf = 1
        for cdf in cdfs:
            pf += cdf * sign
            sign *= -1
        
        #print(pf)
        return pf, "series calculation"
        
    def _get_x(self, n=10, steps=10):
        ## log(np.pi/4) == -x - s + log(x)
        ## log(rad) + s == -x + log(x)
        ## x-log(x) = -log(rad) - s
        
        #rad = exp(-x-s)*x
        #log(rad/x) = -x -s
        #x = -log(rad/x) - s
        
        # d_rad/dx = -exp(-x-s)*x + exp(-x-s)
        # d_rad/dx = exp(-x-s) * (1-x)
        
        ## d2/dx2 = exp(-x-s) * (x-1) - exp(-x-s)
        ## d2/dx2 = exp(-x-s) * (x-2)
        
        rad_max = np.exp(-self._s - 1)
        #print(rad_max)
        k_max = np.floor(rad_max / np.pi - 1/2)
        k = k_max
        #x = f.marginals[0].ppf(1e-200)
        rad = (k - n - 1 + 1/2) * np.pi
        #x = rad / np.exp(-self._s)
        x = -np.log(abs(rad)) - self._s
        xs = []
        
        for i in range(-n, 2):
            #print(rad)
            for __ in range(steps): # 10 out to be enough for everybody
                x = x + (rad  / np.exp(-x-self._s) - x) / (1-x)
                #print(x)
            xs.append(x)
            rad = (k+i + 1/2) * np.pi
        return xs, rad_max
        
    # Fence off!
    def get_2D_R_boundary(self, nrod=100, xlim=(-5,5), ylim=(-5,5)):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        # it is not mine
        #xa = -4.05229846333861
        #xb = -4.95067172463682
        #xc = -5.37859367619679
        #xd = -5.66345816541508
        #xe = -5.87765022259327
        #xf = -6.04950202015156
        #xg = -6.19309680892552
        
        # ikska
        #xes = (xa, xb, xc, xd, xe, xf, xg)
        xes, __rad_max = self._get_x()
        
        boundaries = []
        ymin, ymax = ylim
        
        for x in xes:
            xbound = np.full(nrod, x)
            ybound = np.linspace(ymin, ymax, nrod, endpoint=True)
            # sample compatible
            # малы транспонировать кароно? Озьы кулэ!
            bound_R = np.vstack((xbound, ybound)).T
            boundaries.append(Ingot(bound_R))
            
        
        return boundaries    
            
        



class FourBranch2D:
    """
    Four branch system
    """
    def __init__(self, k1=3, k2=7):
        self.k1 = k1
        self.k2 = k2
        
    # sign
    def __repr__(self):
        return 'FourBranch2D(k1=%s, k2=%s)' % (self.k1, self.k2)
        
    def __call__(self, input_sample):
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample)

        k1, k2 = self.k1, self.k2
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 2)
        x1, x2 = sample[:,0], sample[:,1]
        g1 = k1 + 0.1*(x1 - x2)**2 - (x1 + x2)/np.sqrt(2)
        g2 = k1 + 0.1*(x1 - x2)**2 + (x1 + x2)/np.sqrt(2)
        g3 = (x1 - x2) + k2/np.sqrt(2)
        g4 = (x2 - x1) + k2/np.sqrt(2) #č byl tu překlep v jednom članku
        g = np.min((g1, g2, g3, g4), axis=0)
        return SampleBox(input_sample, g, repr(self))
        
    # Fence off!
    def get_2D_R_boundary(self, nrod=100, *args, **kwargs):
        """
        Fence off!
        nrod - number of rods in fencing
        """
        
        k1, k2 = self.k1, self.k2
        sqrt2 = np.sqrt(2)
        
        m = k1/sqrt2 + k2**2/20/sqrt2
        x1_1 = -m - k2/2/sqrt2
        x1_2 = m - k2/2/sqrt2
        
        
        #č nrod je na každou větev
        xbound_1 = np.linspace(x1_1, x1_2, nrod, endpoint=False)
        ybound_1 = xbound_1 + k2/sqrt2
        
        
        #č naša milovaná parabolka v pootočených souřadnicích
        xbound_2_ = np.linspace(-k2/2, k2/2, nrod, endpoint=False)
        ybound_2_ = xbound_2_**2/5 + k1
        
        xbound_2 = (xbound_2_ + ybound_2_) / sqrt2
        ybound_2 = (-xbound_2_ + ybound_2_) / sqrt2
        
        xbound = np.hstack((xbound_1, xbound_2, -xbound_1, -xbound_2))
        ybound = np.hstack((ybound_1, ybound_2, -ybound_1, -ybound_2))
        
        
        # sample compatible
        # малы транспонировать кароно? Озьы кулэ!
        bound_R = np.array((xbound, ybound)).T
        # tuple of samples
        return (Ingot(bound_R),) 



class PassiveVehicleSuspension:
    """
    """
    def __init__(self, V=10, A=0.15915, b_0=0.27, m=0.8158, M=3.2633, g=981):
        #self.constants = [V, A/2/np.pi, b_0, m, M, g]
        self.constants = [V, A, b_0, m, M, g]
        
    # sign
    def __repr__(self):
        return 'PassiveVehicleSuspension(%s, %s, %s, %s, %s, %s)' % (*self.constants,)
        
    def __call__(self, input_sample):
        V, A, b_0, m, M, g = self.constants
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, 3)
        c, c_k, k = sample[:,0], sample[:,1], sample[:,2]
        K1 = (np.pi * m * V * A) / (b_0 * k * g**2)
        C1 = c_k / (m + M)
        C2 = c / M
        C3 = c**2 / (m * M)
        C4 = c_k * k**2 /(m * M**2)
        G1 = 1 - K1 * ((C1-C2)**2 + C3 + C4)
        #G2 = 1 - 7.6394 / (4000 * (M*g)**(-1.5) * c - 1)
        G2 = 4000 * (M*g)**(-1.5) * c - 8.6394 
        #G3 = 1 - 0.5 / np.sqrt(M * g)  / np.sqrt(k**2 * c_k / c / (M + m) + c)
        G3 = 2* np.sqrt(M * g)  * np.sqrt(k**2 * c_k / c / (M + m) + c) - 1
        G4 = c_k  - (g * (M + m))**0.877 
        print(G1, G2, G3, G4)
        return SampleBox(input_sample, np.min((G1, G2, G3, G4), axis=0), repr(self))



#
# Free functions
#


def quadratic(input_sample):
    selfnvar = 2
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = get_R_coordinates(input_sample, selfnvar)
    x1, x2 = sample[:,0], sample[:,1]
    g = 3 - x1**4/33 - x2
    return SampleBox(input_sample, g, 'quadratic')

def quadratic_R_boundary(nrod=210, xlim=(-5,5), *args):
        boundaries = []
        x_min, x_max = xlim
        x1 = np.linspace(x_min, x_max, nrod, endpoint=True)
        x2 = 3 - x1**4/33
        
        bound_R_1 = np.vstack(( x1, x2)).T
        boundaries.append(Ingot(bound_R_1))
        return boundaries    

quadratic.get_2D_R_boundary = quadratic_R_boundary



# AK-MCS An active learning reliability method combining kriging
def modified_rastrigin(input_sample):
    selfnvar = 2
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = get_R_coordinates(input_sample, selfnvar)
    
    g = 10 - np.sum(sample**2 - 5 * np.cos(2 * np.pi * sample), axis=1)
    return SampleBox(input_sample, g, 'modified_rastrigin')


# Papaioannou. Sequental importance sampling
# for structural reliability analysis
def noisy_lsf(input_sample):
    selfnvar = 6
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = get_R_coordinates(input_sample, selfnvar)
    
    g = np.sum(sample * [1, 2, 2, 1, -5, -5] + 0.001 * np.sin(100 * sample), axis=1)
    return SampleBox(input_sample, g, 'noisy_lsf')

def single_spring_nonlinear_oscilator(input_sample):
    selfnvar = 5
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = get_R_coordinates(input_sample, selfnvar)
    m, c, r, F1, t1 = sample.T
    # I don't care about physical meaningfullness
    # and I don't care if frequency is less than zero
    w0 = np.sqrt(c / m)
    g = 3 * r - np.abs(2 * F1 / m / w0**2 * np.sin(w0 * t1 / 2))
    return SampleBox(input_sample, g, 'single_spring_nonlinear_oscilator')
        
# DOI:10.1016/j.strusafe.2004.11.001
# Beneﬁt of splines and neural networks in simulation based
# structural reliability analysis
def nonlinear_oscilator(input_sample):
    selfnvar = 6
    # očekávam, že get_R_coordinates mně vrátí 2D pole
    sample = get_R_coordinates(input_sample, selfnvar)
    m, c1, c2, r, F1, t1 = sample.T
    # I don't care about physical meaningfullness
    # and I don't care if frequency is less than zero
    w0 = np.sqrt((c1 + c2) / m)
    g = 3 * r - np.abs(2 * F1 / m / w0**2 * np.sin(w0 * t1 / 2))
    return SampleBox(input_sample, g, 'nonlinear_oscilator')    
        
        
def piecewise_2D_linear(input_sample):
        selfnvar = 2
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, selfnvar)
        x1, x2 = sample[:,0], sample[:,1]
        g1 = np.where(x1 > 3.5, 4-x1, 0.85-0.1*x1)
        g2 = np.where(x2 > 2, 0.5-0.1*x2, 2.3-x2)
        g = np.min((g1,g2), axis=0)
        return SampleBox(input_sample, g, 'piecewise_2D_linear')

# boundary
piecewise_2D_linear.get_2D_R_boundary = GetQuadrantBoundary2D(center_point=(4,5), quadrant='III')
piecewise_2D_linear.pf_expression = lambda fm, a=4, b=5: (fm.marginals[0].sf(a) + fm.marginals[1].sf(b) - fm.marginals[0].sf(a)*fm.marginals[1].sf(b), 'exact solution')


       
        
def non_chi_squares(input_sample):
        selfnvar = 2
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, selfnvar)
        x1, x2 = sample[:,0], sample[:,1]
        g = 0.1 * (52 - 1.5 * x1**2 - x2**2)
        return SampleBox(input_sample, g, 'non_chi_squares')


# boundary for non_chi_squares (Breitung with pareto tail)
def non_chi_squares_R_boundary(nrod=210, *args):
     
        boundaries = []
        y = np.linspace(-np.sqrt(52), np.sqrt(52), nrod, endpoint=True)
        x = + np.sqrt(2*(52-y**2)/3)
        bound_R_1 = np.vstack(( x, y)).T
        bound_R_2 = np.vstack((-x, y)).T
        boundaries.append(Ingot(bound_R_1))
        boundaries.append(Ingot(bound_R_2))
        return boundaries    

non_chi_squares.get_2D_R_boundary = non_chi_squares_R_boundary







def branin_2D(input_sample):
        """
        Rescaled Branin function
        """
        selfnvar = 2
        # očekávam, že get_R_coordinates mně vrátí 2D pole
        sample = get_R_coordinates(input_sample, selfnvar)
        x1, x2 = sample[:,0], sample[:,1]
        g = 80 - ((15*x2 - 5/(4*np.pi**2)*(15*x1-5)**2 + 5/np.pi*(15*x1-5)-6)**2 + 10*(1-1/8/np.pi)*np.cos(15*x1-5) + 10)
        return SampleBox(input_sample, g, 'branin_2D')



def neverfall(input_sample):
    g = np.full(len(input_sample), 1, dtype=np.int8)
    return SampleBox(input_sample, g, 'neverfall')





#!/usr/bin/env python
# coding: utf-8

"""
We will prepare here WhiteBox instances
of different well-known 2D problems
"""

import numpy as np
from .. import g_models as gm
from .. import f_models
from ..whitebox import WhiteBox
import scipy.stats as stats
from scipy import integrate # for Pareto tail

__all__ = [
            'uniform_branin_2D',
            'snorm_four_branch_2D',
            'snorm_four_branch_2D_2',
            'snorm_piecewise_2D_linear',
            'piecewise_pareto_tail',
            'snorm_min_2D_linear',
            'snorm_min_2D_logistic',
            'modified_rastrigin',
            'quadratic'
            ]

f = f_models.SNorm(2)
# Uniform-uniform
# from scipy: "In the standard form, the distribution is uniform on [0, 1]."
u = f_models.UnCorD((stats.uniform, stats.uniform))



"""
I would like to use function definitions to clearly isolate ones testcases
"""

# Rescaled Branin function
def uniform_branin_2D():
    wt = WhiteBox(u, gm.branin_2D)
    wt.pf_exact = 0.257
    wt.pf_exact_method = 'known value' #"some guys said me that" it's 0.256
    wt.description = "Rescaled Branin function"
    return wt

# Four branch system
def snorm_four_branch_2D():
    wt = WhiteBox(f, gm.FourBranch2D(k1=3, k2=7))
    wt.pf_exact = 2.34e-03
    wt.pf_exact_method = 'known value' #"some guys said me that" 
    wt.description = "Four branch system from some paper"
    return wt


def snorm_four_branch_2D_2():
    wt = WhiteBox(f, gm.FourBranch2D(k1=5.5, k2=11))
    # TODO calculate
    #wt.pf_exact = 0.257
    #wt.pf_exact_method = 'known value' #"some guys said me that" it's 0.256
    wt.description = "Four branch system from another paper"
    return wt





# Breitung
# Piecewise linear function
def snorm_piecewise_2D_linear():
    wt = WhiteBox(f, gm.piecewise_2D_linear)
    wt.description = "Breitung. Piecewise linear function"
    return wt



# Pareto tail
class PiecewiseParetoDist:
    def __init__(self, a=3.5):
        self.ax = a
        self.au = stats.norm.cdf(a) 
        self.c = np.log(stats.norm.cdf(-a))/np.log(a)
        self.pareto = stats.pareto(b=-self.c)
        # без излишевст
        self._p = integrate.quad(lambda x: self.pdf(x), -np.inf, np.inf)[0]
        self._mean = integrate.quad(lambda x: x*self.pdf(x), -np.inf, np.inf)[0]
        self._var = integrate.quad(lambda x: x**2*self.pdf(x), -np.inf, np.inf)[0]
        
    def cdf(self, x):
        return np.where(x < self.ax, stats.norm.cdf(x), self.pareto.cdf(x))
    def pdf(self, x):
        return np.where(x < self.ax, stats.norm.pdf(x), self.pareto.pdf(x))
    def ppf(self, u):
        return np.where(u < self.au, stats.norm.ppf(u), self.pareto.ppf(u))
        
    def mean(self): return self._mean
    def var(self): return self._var
    def std(self): return np.sqrt(self._var)
    


def piecewise_pareto_tail():
    h = f_models.UnCorD((stats.norm, PiecewiseParetoDist()))
    wt = WhiteBox(h, gm.non_chi_squares)
    wt.pf_exact = 1.87e-06
    wt.pf_exact_method = 'FORM-type thought-three beta points'
    wt.description = "Breitung. Pareto tail"
    return wt



# Logistic 2D function
def snorm_min_2D_linear():
    wt = WhiteBox(f, gm.Logistic2D())
    wt.description = "Breitung. 2D linear (easy version for SuS)"
    return wt

def snorm_min_2D_logistic():
    wt = WhiteBox(f, gm.Logistic2D(easy_version=False))
    wt.description = "Breitung. Logistic 2D function (hard version for SuS)"
    return wt

def modified_rastrigin():
    wt = WhiteBox(f, gm.modified_rastrigin)
    wt.description = "Modified Rastrigin"
    wt.pf_exact = 7.34e-02
    wt.pf_exact_method = 'known value'
    return wt

def quadratic():
    wt = WhiteBox(f, gm.quadratic)
    return wt
    
    



